# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015-2020 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from unittest.mock import patch

from django.core.signals import request_finished
from django.db.models import Q
from django.test import override_settings

from rolepermissions.checkers import has_role

from apps.challenge.tests.factories import (
    QualificationActivityFactory,
    QualificationFactory,
    SessionFactory,
)
from apps.orga.tests.factories import OrganizationFactory
from apps.user import FORMATION_M1, FORMATION_M2
from apps.user.models import UserProfile
from apps.user.tests.factories import UserFactory


def test_str_representations(db):
    user = UserFactory()

    assert str(user) == user.username
    assert str(user.profile) == user.get_full_name()


def test_get_users_that_worked_in_cantons_filters_by_year_and_role(db):
    canton = "VD"
    other_canton = "GE"
    year = 2019
    session = SessionFactory(
        day=datetime.date(year, 4, 10),
        orga=OrganizationFactory(address_canton=canton),
    )
    actor = UserFactory()
    helper = UserFactory()
    leader = UserFactory()
    QualificationFactory(session=session, actor=actor, helpers=[helper], leader=leader)

    # Users working in a different canton or year should not be returned
    QualificationFactory(
        session=SessionFactory(
            day=datetime.date(year + 1, 5, 10),
            orga=OrganizationFactory(address_canton=canton),
        ),
        actor=UserFactory(),
    )
    QualificationFactory(
        session=SessionFactory(
            day=datetime.date(year, 6, 10),
            orga=OrganizationFactory(address_canton=other_canton),
        ),
        actor=UserFactory(),
    )

    users = UserProfile.get_users_that_worked_in_cantons([canton], year=year)
    assert set(users) == {actor, helper, leader}

    extra_user = UserFactory()
    users_with_extra = UserProfile.get_users_that_worked_in_cantons(
        [canton], year=year, additional_or=[Q(pk=extra_user.pk)]
    )
    assert extra_user in users_with_extra


def test_automatic_collaborator_role_assignment_formation(db):
    user = UserFactory()
    user.profile.formation = FORMATION_M1
    user.profile.save()
    assert has_role(user, "collaborator")

    user.profile.formation = ""
    user.profile.save()
    assert not has_role(user, "collaborator")


def test_automatic_collaborator_role_assignment_actor(db):
    user = UserFactory()
    category_c = QualificationActivityFactory(category="C")
    category_c.save()
    user.profile.formation = ""
    user.profile.actor_for.add(category_c)
    user.profile.save()
    assert has_role(user, "collaborator")

    user.profile.actor_for.clear()
    user.profile.save()
    assert not has_role(user, "collaborator")


def test_automatic_collaborator_role_assignment_both(db):
    user = UserFactory()
    category_c = QualificationActivityFactory(category="C")
    category_c.save()
    user.profile.formation = FORMATION_M2
    user.profile.actor_for.add(category_c)
    user.profile.save()
    assert has_role(user, "collaborator")

    user.profile.actor_for.clear()
    user.profile.save()
    # Still has a formation
    assert has_role(user, "collaborator")

    user.profile.formation = ""
    user.profile.save()
    assert not has_role(user, "collaborator")


@patch("apps.user.signals.send_mail")
@override_settings(PROFILE_CHANGED_NOTIFY_EMAIL="admin@example.com")
@override_settings(DEFAULT_FROM_EMAIL="system@example.com")
def test_userprofile_field_change_signal_iban_change(mock_send_mail, db, settings):
    mock_send_mail.reset_mock()

    user = UserFactory(profile__iban="CH9300762011623852957")

    user.profile.iban = "CH1404835012345678009"
    user.profile.save()

    # Trigger the end of the request to send the email
    request_finished.send(sender=None)

    mock_send_mail.assert_called_once()
    args = mock_send_mail.call_args[0]
    assert "Notification de modification de données utilisateur·trice" in args[0]
    assert (
        f"Les informations suivantes pour l’utilisateur·trice {user.get_full_name()} / {user.pk} ont été modifiées"
        in args[1]
    )
    assert "IBAN: CH93 0076 2011 6238 5295 7 → CH14 0483 5012 3456 7800 9" in args[1]
    assert args[2] == "system@example.com"
    assert args[3] == ["admin@example.com"]


@patch("apps.user.signals.send_mail")
@override_settings(PROFILE_CHANGED_NOTIFY_EMAIL="admin@example.com")
@override_settings(DEFAULT_FROM_EMAIL="system@example.com")
def test_userprofile_field_change_signal_address_changes(mock_send_mail, db, settings):
    mock_send_mail.reset_mock()

    user = UserFactory(
        profile__address_street="Old Street",
        profile__address_no="123",
        profile__address_zip="1000",
        profile__address_city="Old City",
        profile__address_canton="VD",
    )

    user.profile.address_street = "New Street"
    user.profile.address_no = "456"
    user.profile.address_city = "New City"
    user.profile.save()

    # Trigger the end of the request to send the email
    request_finished.send(sender=None)

    mock_send_mail.assert_called_once()
    args = mock_send_mail.call_args[0]
    assert "Notification de modification de données utilisateur·trice" in args[0]
    assert (
        f"Les informations suivantes pour l’utilisateur·trice {user.get_full_name()} / {user.pk} ont été modifiées"
        in args[1]
    )

    # Check that all three address changes are in the email body
    assert "address_street: Old Street → New Street" in args[1]
    assert "address_no: 123 → 456" in args[1]
    assert "address_city: Old City → New City" in args[1]

    assert args[2] == "system@example.com"
    assert args[3] == ["admin@example.com"]


@patch("apps.user.signals.send_mail")
def test_userprofile_field_change_signal_no_changes_when_empty_old_values(
    mock_send_mail, db, settings
):
    mock_send_mail.reset_mock()

    user = UserFactory()
    user.profile.save()

    user.profile.iban = "CH93 0076 2011 6238 5295 7"
    user.profile.address_street = "New Street"
    user.profile.save()

    mock_send_mail.assert_not_called()


@patch("apps.user.signals.send_mail")
def test_userprofile_field_change_signal_no_changes_when_values_same(
    mock_send_mail, db, settings
):
    mock_send_mail.reset_mock()
    user = UserFactory(
        profile__iban="CH93 0076 2011 6238 5295 7",
        profile__address_street="Same Street",
    )

    user.profile.iban = "CH93 0076 2011 6238 5295 7"
    user.profile.address_street = "Same Street"
    user.profile.save()
    mock_send_mail.assert_not_called()


@patch("apps.user.signals.send_mail")
@override_settings(PROFILE_CHANGED_NOTIFY_EMAIL="admin@example.com")
@override_settings(DEFAULT_FROM_EMAIL="system@example.com")
def test_userprofile_unified_post_save_notification_sends_email(
    mock_send_mail, db, settings
):
    """
    Test that when multiple models (user + profile) are changed in a single save, the mail is sent only once.
    """
    mock_send_mail.reset_mock()

    user = UserFactory(
        email="mytest@example.com",
        profile__iban="CH93 0076 2011 6238 5295 7",
        profile__address_street="Same Street",
    )

    user.profile.iban = "CH14 0483 5012 3456 7800 9"
    user.profile.address_street = "Different Street"
    user.email = "changed@example.com"
    user.profile.save()
    user.save()

    # Trigger the end of the request to send the email
    request_finished.send(sender=None)

    mock_send_mail.assert_called_once()
    args = mock_send_mail.call_args[0]
    assert "Notification de modification de données utilisateur·trice" in args[0]
    assert (
        f"Les informations suivantes pour l’utilisateur·trice {user.get_full_name()} / {user.pk} ont été modifiées"
        in args[1]
    )
    assert "IBAN: CH93 0076 2011 6238 5295 7 → CH14 0483 5012 3456 7800 9" in args[1]
    assert "address_street: Same Street → Different Street" in args[1]
    assert "email: mytest@example.com → changed@example.com" in args[1]



# Generated by Django 1.11.4 on 2017-08-18 09:45
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0030_auto_20170817_1642"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="status",
            field=models.PositiveSmallIntegerField(
                choices=[
                    (0, "---------"),
                    (10, "Actif"),
                    (20, "Réserve"),
                    (30, "Inactif"),
                    (40, "Archive"),
                    (99, "Supprimé"),
                ],
                default=0,
                verbose_name="Statut",
            ),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2020 Didier 'OdyX' Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db.models import Count
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView
from django.views.generic.dates import YearMixin
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView

from rolepermissions.mixins import HasPermissionsMixin

from apps.challenge.export import InvoiceResource
from apps.orga.models import Organization
from defivelo.roles import has_permission, user_cantons

from ...common.views import ExportMixin
from ..forms import InvoiceForm, InvoiceFormQuick
from ..models import Invoice
from .mixins import CantonSeasonFormMixin
from .season import SeasonListView, SeasonMixin


def user_can_edit_invoice(user, invoice: Invoice = None, locked: bool = False):
    if locked or invoice.is_locked:
        return has_permission(user, "challenge_invoice_reset_to_draft")
    else:
        return has_permission(user, "challenge_invoice_cru")


class InvoiceMixin(CantonSeasonFormMixin, HasPermissionsMixin):
    model = Invoice
    required_permission = "challenge_invoice_cru"
    context_object_name = "invoice"
    form_class = InvoiceForm
    raise_without_cantons = True

    def dispatch(self, request, *args, **kwargs):
        self.organization = get_object_or_404(
            Organization, pk=self.kwargs.get("orgapk")
        )
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(organization=self.organization, season=self.season)
        )

    def get_form_kwargs(self, **kwargs):
        kw = super().get_form_kwargs(**kwargs)
        # Remove cantons, we don't need them.
        kw.pop("cantons")
        kw["organization"] = self.organization
        return kw

    def get_initial(self, **kwargs):
        kw = super().get_initial(**kwargs)
        kw["organization"] = self.organization
        kw["season"] = self.season
        return kw

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Add our menu_category context
        context["menu_category"] = "season"
        context["season"] = self.season
        context["organization"] = self.organization
        return context

    def get_object(self):
        invoiceref = self.kwargs.get("invoiceref")

        return get_object_or_404(
            self.model.objects.prefetch_related("lines"),
            season=self.season,
            organization=self.organization,
            ref=invoiceref,
        )


class InvoiceDetailView(InvoiceMixin, DetailView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Add our menu_category context
        invoice = self.get_object()
        lines_per_day = dict()

        for line in invoice.lines.all():
            date_of_current_line = line.historical_session.day

            if date_of_current_line in lines_per_day:
                lines_per_day[date_of_current_line]["line_sum_nb_participants"] = (
                    lines_per_day[date_of_current_line]["line_sum_nb_participants"]
                    + line.nb_participants
                )
                lines_per_day[date_of_current_line]["line_sum_cost_participants"] = (
                    lines_per_day[date_of_current_line]["line_sum_cost_participants"]
                    + line.cost_participants
                )

                lines_per_day[date_of_current_line]["max_nb_bikes"] = max(
                    [line.nb_bikes, lines_per_day[date_of_current_line]["max_nb_bikes"]]
                )

                lines_per_day[date_of_current_line]["line_sum_nb_of_bikes"] = (
                    lines_per_day[date_of_current_line]["line_sum_nb_of_bikes"]
                    + line.nb_bikes
                )

                lines_per_day[date_of_current_line]["max_cost_bikes"] = max(
                    [
                        line.cost_bikes,
                        lines_per_day[date_of_current_line]["max_cost_bikes"],
                    ]
                )

                lines_per_day[date_of_current_line]["line_total"] = (
                    lines_per_day[date_of_current_line]["line_total"]
                    + line.cost_participants
                    + line.cost_bikes_reduced
                )

                lines_per_day[date_of_current_line]["max_cost_bikes_reduced"] = max(
                    [
                        line.cost_bikes_reduced,
                        lines_per_day[date_of_current_line]["max_cost_bikes_reduced"],
                    ]
                )

            else:
                lines_per_day.update(
                    {
                        date_of_current_line: {
                            "obj": line,
                            "line_sum_nb_participants": line.nb_participants,
                            "max_nb_bikes": line.nb_bikes,
                            "line_sum_nb_of_bikes": line.nb_bikes,
                            "max_cost_bikes": line.cost_bikes,
                            "line_total": line.cost_bikes_reduced
                            + line.cost_participants,
                            "line_sum_cost_participants": line.cost_participants,
                            "max_cost_bikes_reduced": line.cost_bikes_reduced,
                        }
                    }
                )

            if line.has_cost_bikes:
                lines_per_day[date_of_current_line]["cost_bikes"] = line.cost_bikes
                if line.cost_bikes_reduction_percent():
                    lines_per_day[date_of_current_line]["cost_bikes_reduced"] = (
                        line.cost_bikes_reduced
                    )
                    lines_per_day[date_of_current_line][
                        "cost_bikes_reduction_percent"
                    ] = line.cost_bikes_reduction_percent

        adjusted_sum_of_bikes = sum(
            [line["max_nb_bikes"] for _, line in lines_per_day.items()]
        )

        context["user_can_edit_invoice"] = user_can_edit_invoice(
            self.request.user, invoice
        )
        context["cost_per_participant"] = invoice.settings.cost_per_participant
        context["cost_per_bike"] = invoice.settings.cost_per_bike
        context["filtered_lines"] = lines_per_day
        context["adjusted_sum_of_bikes"] = adjusted_sum_of_bikes
        context["any_line_has_reduction"] = any(
            [
                line["obj"].cost_bikes_reduction_percent()
                for line in lines_per_day.values()
            ]
        )
        if not invoice.is_locked and not invoice.is_up_to_date:
            context["refresh_form"] = InvoiceFormQuick(instance=invoice)

        context["pdf_file_name"] = _("Facture") + "_" + self.object.ref + "_DEFIVELO"

        return context


class InvoiceCreateView(InvoiceMixin, CreateView):
    pass


class InvoiceListView(InvoiceMixin, ListView):
    context_object_name = "invoices"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["user_can_edit_locked_invoice"] = user_can_edit_invoice(
            user=self.request.user, invoice=None, locked=True
        )
        return context


class InvoiceYearlyListView(SeasonListView, HasPermissionsMixin, ListView):
    required_permission = "challenge_invoice_cru"
    template_name = "challenge/invoice_yearly_list.html"
    allow_empty = True
    allow_future = True
    make_object_list = True

    def get_context_data(self, **kwargs):
        """
        Add the needed context to set the menu and submenu categories correctly
        """
        context = super().get_context_data(**kwargs)
        context["menu_category"] = "finance"
        context["submenu_category"] = "invoices"
        return context

    def get_queryset(self, **kwargs):
        return (
            super()
            .get_queryset(**kwargs)
            .filter(cantons__overlap=user_cantons(self.request.user))
            .prefetch_related("invoices", "invoices__lines", "invoices__organization")
            .annotate(nb_invoices=Count("invoices", distinct=True))
        )


class InvoiceListExport(ExportMixin, YearMixin, ListView):
    export_class = InvoiceResource()
    export_filename = _("Invoices")
    required_permission = "export_yearly_invoices"

    def get_queryset(self):
        queryset = Invoice.objects.filter(
            season__year=self.get_year(), status=Invoice.STATUS_VALIDATED
        ).order_by("generated_at")
        return queryset


class InvoiceUpdateView(InvoiceMixin, UpdateView):
    def get_form(self, *args, **kwargs):
        """
        Disallow any field edition if the invoice is locked
        """
        form = super().get_form(*args, **kwargs)
        if not user_can_edit_invoice(self.request.user, self.get_object()):
            for field in form.fields:
                form.fields[field].disabled = True
        return form


class SeasonOrgaListView(SeasonMixin, HasPermissionsMixin, ListView):
    context_object_name = "organizations"
    required_permission = "challenge_invoice_cru"

    def get_queryset(self):
        return (
            Organization.objects.filter(sessions__in=self.season.sessions_with_qualifs)
            .annotate(nb_invoices=Count("invoices", distinct=True))
            .distinct()
        )



from django import template

register = template.Library()


@register.filter
def field_type(field):
    """
    Get the name of the field class.
    """
    if hasattr(field, "field"):
        field = field.field
    s = str(type(field.widget).__name__)
    s = s.lower()
    return s



from datetime import date

from django.conf import settings
from django.contrib import messages
from django.contrib.sites.models import Site
from django.core import mail
from django.core.exceptions import PermissionDenied
from django.db.models import Count, F, Q
from django.db.models.functions import Upper
from django.shortcuts import get_object_or_404, redirect
from django.template.defaultfilters import date as datefilter
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import formats, timezone, translation
from django.utils.datastructures import OrderedSet
from django.utils.dates import MONTHS_3
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext as n
from django.views.generic import RedirectView, TemplateView
from django.views.generic.dates import MonthArchiveView
from django.views.generic.edit import FormView

from localflavor.ch.ch_states import STATE_CHOICES
from tablib import Dataset

from apps.challenge.models.session import Session
from apps.common import DV_STATE_CHOICES
from apps.common.views import ExportMixin
from apps.salary import BONUS_LEADER, HOURLY_RATE_HELPER, RATE_ACTOR
from apps.salary.forms import ControlTimesheetFormSet, TimesheetFormSet
from apps.salary.models import Timesheet
from defivelo.roles import has_permission, user_cantons

from ...user.models import UserProfile
from ...user.views.standard import ReturnUrlMixin
from .. import timesheets_overview


class RedirectUserMonthlyTimesheets(RedirectView):
    is_permanent = False

    def get_redirect_url(self, *args, **kwargs):
        kwargs["pk"] = self.request.user.pk
        return reverse(
            "salary:user-timesheets",
            kwargs=kwargs,
        )


class UserMonthlyTimesheets(MonthArchiveView, ReturnUrlMixin, FormView):
    date_field = "day"
    month_format = "%m"
    allow_empty = True
    allow_future = True
    template_name = "salary/month_timesheets.html"

    @property
    def success_url(self):
        return reverse("salary:timesheets-overview", kwargs={"year": self.get_year()})

    def get_queryset(self):
        return (
            (
                Session.objects.values("day")
                .filter(
                    Q(qualifications__actor=self.selected_user)
                    | Q(qualifications__helpers=self.selected_user)
                    | Q(qualifications__leader=self.selected_user)
                )
                .filter(day__lte=timezone.now())
                .annotate(
                    orga_count=Count("orga_id", distinct=True),
                    orga_canton=Upper(F("orga__address_canton")),
                    helper_count=Count(
                        "qualifications__pk",
                        filter=Q(qualifications__helpers=self.selected_user)
                        | Q(qualifications__leader=self.selected_user),
                        distinct=True,
                    ),
                    leader_count=Count(
                        "qualifications__pk",
                        filter=Q(qualifications__leader=self.selected_user),
                        distinct=True,
                    ),
                    actor_count=Count(
                        "qualifications__pk",
                        filter=Q(qualifications__actor=self.selected_user),
                        distinct=True,
                    ),
                )
            )
            .exclude(actor_count=0, helper_count=0)
            .order_by("day")
        )

    def get_form(self, form_class=None):
        """Return an instance of the form to be used in this view."""
        form = super().get_form(form_class=form_class)

        # Return the form as-is if the user is not validating timesheets
        if not has_permission(self.request.user, "timesheet_editor"):
            return form

        # We disable forms where the session is not in the managed canton.
        def disable_form(form, reason=None):
            for key in form.fields:
                form.fields[key].disabled = True
            form.disabled = True
            setattr(form, "disabled_reason", reason)

        # A validator editing their own hours is treated like any monitor:
        # the canton restriction only applies when validating other people's hours.
        if self.request.user.pk == self.selected_user.pk:
            return form

        user_cantons_list = [str(c).upper() for c in user_cantons(self.request.user)]
        for i, session in enumerate(self.object_list):
            session_canton = session.get("orga_canton", None)
            session_canton_label = dict(STATE_CHOICES).get(
                session_canton, session_canton
            )

            if session_canton and session_canton not in user_cantons_list:
                disable_form(
                    form.forms[i],
                    _(
                        'La session a lieu dans le canton "%s" que vous n\'administrez pas.'
                    )
                    % session_canton_label,
                )
                setattr(form, "has_disabled_forms", True)

        return form

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["year"] = self.get_year()
        context["menu_category"] = ["timesheet"]
        context["monitor_name"] = self.selected_user.get_full_name()
        namespaces = self.request.resolver_match.namespaces
        url_name = self.request.resolver_match.url_name
        context["prev_url"] = reverse(
            ":".join([*namespaces, url_name]),
            kwargs={
                "month": context["previous_month"].month,
                "year": context["previous_month"].year,
                "pk": self.selected_user.id,
            },
        )

        context["next_url"] = reverse(
            ":".join([*namespaces, url_name]),
            kwargs={
                "month": context["next_month"].month,
                "year": context["next_month"].year,
                "pk": self.selected_user.id,
            },
        )
        context["formset"] = context["form"]

        # Consider a field "visible" if it is visible in at least one of the forms
        visible_fields = OrderedSet()
        for form in context["form"].forms:
            for field in form.visible_fields():
                visible_fields.add(field)

        context["fields_grouped_by_field_name"] = (
            {
                fieldname.label: [
                    form[fieldname.name] for form in context["form"].forms
                ]
                for fieldname in visible_fields
            }
            if context["form"].forms
            else {}
        )
        context["can_print"] = all(
            form.initial.get("validated") for form in context["form"].forms
        )
        context["in_the_future"] = date.today() < context["month"]
        context["is_current_month"] = date.today().replace(day=1) == context["month"]
        all_sessions = (
            Session.objects.filter(
                Q(qualifications__actor=self.selected_user)
                | Q(qualifications__helpers=self.selected_user)
                | Q(qualifications__leader=self.selected_user)
            )
            .filter(day__in=[o["day"] for o in self.object_list])
            .prefetch_related("orga")
            .distinct()
        )
        context["all_sessions_by_day"] = {
            o["day"]: [s for s in all_sessions if s.day == o["day"]]
            for o in self.object_list
        }
        context["orphaned_timesheets"] = (
            Timesheet.objects.filter(
                user=self.selected_user,
                date__year=context["year"],
                date__month=context["month"].month,
            )
            .exclude(date__in=[o["day"] for o in self.object_list])
            .order_by("date")
        )
        return context

    def get_month(self):
        return super().get_month() or date.today().month

    def get_year(self):
        return super().get_year() or date.today().year

    def get_initial(self):
        initial = []
        timesheets = {
            timesheet.date: timesheet
            for timesheet in Timesheet.objects.filter(
                date__in=[obj["day"] for obj in self.object_list],
                user=self.selected_user,
            )
        }
        for session in self.object_list:
            timesheet = timesheets.get(session["day"])
            attributes = {
                "date": session["day"],
                "time_helper": session["helper_count"]
                * (
                    4
                    if session["orga_count"] == 1 and session["helper_count"] > 1
                    else 4.5
                ),
                "actor_count": session["actor_count"],
                "leader_count": session["leader_count"],
                "overtime": timesheet.overtime if timesheet else 0,
                "traveltime": timesheet.traveltime if timesheet else 0,
                "validated": bool(timesheet.validated_at) if timesheet else False,
                "ignore": timesheet.ignore if timesheet else False,
                "comments": timesheet.comments if timesheet else "",
            }
            initial.append(attributes)
        return initial

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["form_kwargs"] = {
            "validator": self.request.user,
            "selected_user": self.selected_user,
        }
        return kwargs

    def form_valid(self, formset):
        """If the form is valid, save the associated model."""
        for form in formset:
            form.save()
        return super().form_valid(form)

    def post(self, request, *args, **kwargs):
        self.date_list, self.object_list, extra_context = self.get_dated_items()
        return super().post(request, *args, **kwargs)

    def form_invalid(self, form):
        """If the form is invalid, render the invalid form."""
        self.date_list, self.object_list, extra_context = self.get_dated_items()
        context = self.get_context_data(
            object_list=self.object_list,
            date_list=self.date_list,
            form=form,
            **extra_context,
        )
        return self.render_to_response(context)

    def dispatch(self, request, *args, **kwargs):
        if not has_permission(self.request.user, "timesheet") and not has_permission(
            self.request.user, "timesheet_editor"
        ):
            raise PermissionDenied
        self.selected_user = (
            get_object_or_404(
                timesheets_overview.get_visible_users(self.request.user),
                pk=self.kwargs["pk"],
            )
            if self.kwargs["pk"]
            else self.request.user
        )
        return super().dispatch(request, *args, **kwargs)

    def get_form_class(self):
        if has_permission(self.request.user, "timesheet_editor"):
            return ControlTimesheetFormSet
        else:
            return TimesheetFormSet


class YearlyTimesheets(TemplateView):
    template_name = "salary/timesheets_yearly_overview.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["menu_category"] = ["timesheet"]

        year = self.kwargs["year"]
        context["canton"] = active_canton = self.request.GET.get("canton")

        users = timesheets_overview.get_visible_users(self.request.user).order_by(
            "first_name", "last_name"
        )
        global_timesheets_status_matrix = (
            timesheets_overview.get_timesheets_status_matrix(year=year, users=users)
        )
        if active_canton:
            users = UserProfile.get_users_that_worked_in_cantons(
                [active_canton], year=year
            )
            timesheets_status_matrix = timesheets_overview.get_timesheets_status_matrix(
                year=year, users=users, cantons=[active_canton]
            )
        else:
            timesheets_status_matrix = global_timesheets_status_matrix

        context["months"] = MONTHS_3
        context["timesheets_status_matrix"] = timesheets_status_matrix
        context["show_reminder_button_months"] = (
            timesheets_overview.get_missing_timesheet_status_per_month(
                global_timesheets_status_matrix
            )
        )
        context["timesheets_amount"] = (
            timesheets_overview.get_timesheets_amount_by_month(
                year=year,
                users=users,
                cantons=[active_canton] if active_canton else None,
            )
        )
        context["orphaned_timesheets"] = (
            timesheets_overview.get_orphaned_timesheets_per_month(
                year=year,
                users=users,
                cantons=[active_canton]
                if active_canton in dict(DV_STATE_CHOICES)
                else None,
            )
        )
        context["cantons"] = DV_STATE_CHOICES
        context["active_canton"] = (
            dict(DV_STATE_CHOICES)[active_canton] if active_canton else None
        )

        return context

    def dispatch(self, request, *args, **kwargs):
        if not has_permission(self.request.user, "timesheet"):
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)


class ExportMonthlyTimesheets(ExportMixin, MonthArchiveView):
    date_field = "date"
    month_format = "%m"
    allow_empty = False
    allow_future = False
    model = Timesheet
    export_kwargs = {"delimiter": ";"}

    def get_queryset(self):
        active_canton = self.request.GET.get("canton")
        users = timesheets_overview.get_visible_users(self.request.user).order_by(
            "first_name", "last_name"
        )
        if active_canton:
            users = users.filter(profile__affiliation_canton=active_canton)
        return super().get_queryset().filter(validated_at__isnull=False, user__in=users)

    def get_dataset_title(self):
        return _("Export Winbiz {month} {year}").format(
            month=self.get_month(), year=self.get_year()
        )

    @property
    def export_filename(self):
        return "%s-%s-%s" % ("export-winbiz", self.get_year(), self.get_month())

    def get_dataset(self, html=False):
        dataset = Dataset()
        _, object_list, _ = self.get_dated_items()

        # See DEFIVELO-218
        category_codes_list = {
            "actor_count": 1106,  # Interventions
            "leader_count": 1105,  # Prime Moniteur·trice 2
            "time_helper": 1101,  # Salaire heures moniteur
            "traveltime": 1102,  # Salaire heures de trajet
            "overtime": 1103,  # Salaire heures supplémentaires
        }

        for salary_details in timesheets_overview.get_salary_details_list(
            object_list
        ).order_by():
            employee_line = [
                datefilter(date.today(), "d.m.Y"),  # Date of the export
                self.get_month(),  # Concerned month
                salary_details["employee_code"],
            ]
            # Add one line to the dataset for every category
            for category, category_code in category_codes_list.items():
                if salary_details[category] != 0:
                    dataset.append(
                        employee_line
                        + [
                            category_code,
                            salary_details[category],
                            "",  # Empty columns, see DEFIVELO-224
                            "",
                            "",
                            "",
                            1,
                            salary_details["first_name"],
                            salary_details["last_name"],
                        ]
                    )

        return dataset


class ExportMonthlyControl(ExportMixin, MonthArchiveView):
    date_field = "date"
    month_format = "%m"
    allow_empty = True
    allow_future = True
    model = Timesheet

    def get_queryset(self):
        active_canton = self.request.GET.get("canton")
        users = timesheets_overview.get_visible_users(self.request.user).order_by(
            "first_name", "last_name"
        )
        if active_canton:
            users = users.filter(profile__affiliation_canton=active_canton)
        return super().get_queryset().filter(validated_at__isnull=False, user__in=users)

    def get_dataset_title(self):
        return _("Export de contrôle {month} {year}").format(
            month=self.get_month(), year=self.get_year()
        )

    @property
    def export_filename(self):
        return f"export-control-{self.get_year()}-{self.get_month()}"

    def get_dataset(self, html=False):
        dataset = Dataset()
        dataset.headers = [
            gettext("Numéro d’employé Crésus"),
            gettext("Prénom"),
            gettext("Nom"),
            gettext("Heures moni·teur·trice ({price}.-/h)").format(
                price=HOURLY_RATE_HELPER
            ),
            gettext("Intervention(s) ({price}.-/Qualif’)").format(price=RATE_ACTOR),
            gettext(
                "Participation(s) comme moni·teur·trice 2 ({price}.-/Qualif’)"
            ).format(price=BONUS_LEADER),
            gettext("Heures supplémentaires ({price}.-/h)").format(
                price=HOURLY_RATE_HELPER
            ),
            gettext("Heures de trajet (aller-retour)"),
            gettext("Total heures"),
            gettext("Total CHF"),
            # gettext("Ne compter aucune heure de travail"),
        ]

        _, object_list, _ = self.get_dated_items()

        for salary_details in timesheets_overview.get_salary_details_list(
            object_list
        ).order_by("first_name"):
            time_total = (
                salary_details["time_helper"]
                + salary_details["overtime"]
                + salary_details["traveltime"]
            )
            chf_total = (
                time_total * HOURLY_RATE_HELPER
                + salary_details["actor_count"] * RATE_ACTOR
                + salary_details["leader_count"] * BONUS_LEADER
            )
            dataset.append(
                [
                    # See DEFIVELO-225
                    salary_details["employee_code"],
                    salary_details["first_name"],
                    salary_details["last_name"],
                    salary_details["time_helper"],
                    salary_details["actor_count"],
                    salary_details["leader_count"],
                    salary_details["overtime"],
                    salary_details["traveltime"],
                    time_total,
                    chf_total,
                ]
            )
        return dataset


class CleanupOrphanedTimesheets(TemplateView):
    template_name = "salary/cleanup_orphaned_timesheets.html"
    required_permission = "timesheet_editor"

    def dispatch(self, request, *args, month, year, **kwargs):
        if not has_permission(request.user, self.required_permission):
            raise PermissionDenied
        self.month = int(month)
        self.year = int(year)

        active_canton = self.request.GET.get("canton")
        self.orphaned_timesheets = (
            timesheets_overview.get_orphaned_timesheets_per_month(
                year=year,
                month=self.month,
                users=timesheets_overview.get_visible_users(self.request.user),
                cantons=(active_canton if active_canton in DV_STATE_CHOICES else None),
            )
        )
        if not self.orphaned_timesheets:
            messages.success(
                request,
                _("Aucune feuille d'heure orpheline à supprimer; tout va bien."),
            )
            return redirect(self.get_redirect_url())
        return super().dispatch(request, *args, **kwargs)

    def get_redirect_url(self):
        return reverse("salary:timesheets-overview", kwargs={"year": self.year})

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["redirect_url"] = self.get_redirect_url()
        context["period"] = formats.date_format(date(self.year, self.month, 1), "F Y")
        context["orphaned_timesheets"] = self.orphaned_timesheets
        return context

    def post(self, request, *args, **kwargs):
        (deleted, _) = Timesheet.objects.filter(
            id__in=[t.id for t in self.orphaned_timesheets]
        ).delete()
        messages.warning(
            request,
            n(
                "{n} feuille d'heures orpheline supprimée.",
                "{n} feuilles d'heures orphelines supprimées.",
                deleted,
            ).format(n=deleted),
        )
        return redirect(self.get_redirect_url())


class SendTimesheetsReminder(TemplateView):
    template_name = "salary/send_timesheets_reminder.html"
    required_permission = "timesheet_editor"

    def dispatch(self, request, *args, month, year, **kwargs):
        if not has_permission(request.user, self.required_permission):
            raise PermissionDenied
        self.month = int(month)
        self.year = int(year)
        users = (
            timesheets_overview.get_visible_users(self.request.user)
            .order_by("first_name", "last_name")
            .select_related("profile")
        )
        self.recipients = timesheets_overview.get_users_with_missing_timesheets(
            self.year, self.month, users
        )
        if not self.recipients:
            messages.warning(
                request,
                _("Aucun rappel à envoyer: les heures de ce mois sont déjà remplies."),
            )
            return redirect(self.get_redirect_url())
        return super().dispatch(request, *args, **kwargs)

    def get_redirect_url(self):
        return reverse("salary:timesheets-overview", kwargs={"year": self.year})

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["redirect_url"] = self.get_redirect_url()
        context["recipients"] = self.recipients
        context["period"] = formats.date_format(date(self.year, self.month, 1), "F Y")
        email_subject, email_text, *_ = self.render_email_for_user()
        context["email_text"] = email_text
        context["email_subject"] = email_subject
        return context

    def post(self, request, *args, **kwargs):
        emails = [self.render_email_for_user(user) for user in self.recipients]
        mail.send_mass_mail(emails)
        messages.success(request, _("Rappel de soumission des heures expédié."))
        return redirect(self.get_redirect_url())

    def render_email_for_user(self, user=None):
        try:
            email_lang = user.profile.language or translation.get_language()
        except AttributeError:
            email_lang = translation.get_language()

        with translation.override(email_lang):
            timesheets_url = self.request.build_absolute_uri(
                reverse(
                    "salary:my-timesheets",
                    kwargs={"year": self.year, "month": self.month},
                )
            )
            message = render_to_string(
                "salary/email_send_timesheets_reminder.txt",
                {
                    "profile": {
                        "get_full_name": f"{user.first_name} {user.last_name}"
                        if user
                        else _("{Prénom} {Nom}")
                    },
                    "current_site": Site.objects.get_current(),
                    "timesheets_url": timesheets_url,
                },
                self.request,
            )
            return (
                settings.EMAIL_SUBJECT_PREFIX + gettext("Soumission des heures"),
                message,
                settings.DEFAULT_FROM_EMAIL,
                [user.email if user else None],
            )



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2017,2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date, timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Count, Q, Sum
from django.template.defaultfilters import date as datefilter
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from tablib import Dataset

from apps.challenge.models.session import Session
from apps.common import DV_SEASON_AUTUMN, DV_SEASON_LAST_SPRING_MONTH, DV_STATE_CHOICES
from apps.orga.models import Organization
from defivelo.roles import user_cantons
from defivelo.templatetags.dv_filters import (
    season_month_end,
    season_month_start,
    season_verb,
)

linktxt = '<a href="{url}">{content}</a>'


class SeasonSessionsMixin(object):
    limit_to_my_cantons = True

    def get_queryset(self):
        begin = date(self.export_year, 1, 1)
        if self.export_season == DV_SEASON_AUTUMN:
            begin = date(self.export_year, DV_SEASON_LAST_SPRING_MONTH + 1, 1)
        end = date(self.export_year, DV_SEASON_LAST_SPRING_MONTH + 1, 1) - timedelta(
            days=1
        )
        if self.export_season == DV_SEASON_AUTUMN:
            end = date(self.export_year, 12, 31)
        sessions = (
            Session.objects.filter(
                day__gte=begin,
                day__lte=end,
            )
            .order_by("day", "begin")
            .prefetch_related("orga")
        )
        if self.limit_to_my_cantons:
            sessions = sessions.filter(
                orga__address_canton__in=user_cantons(self.request.user)
            )
        return sessions


class QualifsCalendarExport(SeasonSessionsMixin):
    def get_dataset_title(self):
        return "Calendar export dataset title"

    @property
    def export_filename(self):
        return "%s-%s-%s" % (
            _("Calendar"),
            self.export_year,
            season_verb(self.export_season),
        )

    def get_dataset(self, html=False):
        dataset = Dataset()
        dataset.headers = [
            gettext("Semaine"),
            gettext("Lundi"),
            gettext("Mardi"),
            gettext("Mercredi"),
            gettext("Jeudi"),
            gettext("Vendredi"),
            gettext("Samedi"),
            gettext("Dimanche"),
        ]

        calendar_data = self.get_context_data()["date_sessions"]

        row = []
        for day_data in calendar_data:
            day = day_data["day"]

            if day.weekday() == 0:
                row.append(day.strftime("%W"))

            text = day.strftime("%Y-%m-%d")

            for session in day_data["sessions"]:
                text += "\n{} {} {}".format(
                    session.begin if session.begin else "",
                    session.orga.abbr if session.orga.abbr else session.orga.name,
                    session.orga.address_canton,
                )
            row.append(text)

            if day.weekday() == 6:
                dataset.append(row)
                row = []

        return dataset


class SeasonStatsExport(SeasonSessionsMixin):
    def get_dataset_title(self):
        return "{title} - {month_start} à {month_end} {year}".format(
            title=(gettext("Statistiques")),
            year=self.export_year,
            month_start=season_month_start(self.export_season),
            month_end=season_month_end(self.export_season),
        )

    @property
    def export_filename(self):
        return "%s-%s-%s" % (
            _("Stats_Mois"),
            self.export_year,
            season_verb(self.export_season),
        )

    def get_dataset(self, html=False):
        dataset = Dataset()
        dataset.headers = [
            gettext("Canton"),
            gettext("Établissements"),
            gettext("Sessions"),
            gettext("Qualifs"),
            gettext("Nombre d’élèves"),
            gettext("Prêts de vélos"),
            gettext("Prêts de casques"),
            gettext("Nombre de personnes ayant exercé"),
            gettext("… comme moniteur·trice·s 2"),
            gettext("… comme moniteur·trice·s 1"),
            gettext("… comme intervenant·e·s"),
        ]
        volunteers = get_user_model().objects
        orgas = Organization.objects
        object_list = self.get_queryset()
        for canton, canton_verb in DV_STATE_CHOICES:
            cantonal_sessions = object_list.filter(
                orga__address_canton=canton
            ).annotate(
                n_qualifs=Count("qualifications"),
                n_participants=Sum("qualifications__n_participants"),
                n_bikes=Sum("qualifications__n_bikes"),
                n_helmets=Sum("qualifications__n_helmets"),
            )
            n_sessions = cantonal_sessions.count()
            if n_sessions > 0:
                sessions_pks = cantonal_sessions.values_list("id", flat=True)
                dataset.append(
                    [
                        canton,
                        orgas.filter(sessions__in=sessions_pks).distinct().count(),
                        n_sessions,
                        cantonal_sessions.aggregate(total=Sum("n_qualifs"))["total"],
                        cantonal_sessions.aggregate(total=Sum("n_participants"))[
                            "total"
                        ],
                        cantonal_sessions.aggregate(total=Sum("n_bikes"))["total"],
                        cantonal_sessions.aggregate(total=Sum("n_helmets"))["total"],
                        volunteers.filter(
                            Q(qualifs_mon2__session__in=sessions_pks)
                            | Q(qualifs_mon1__session__in=sessions_pks)
                            | Q(qualifs_actor__session__in=sessions_pks)
                        )
                        .distinct()
                        .count(),
                        volunteers.filter(qualifs_mon2__session__in=sessions_pks)
                        .distinct()
                        .count(),
                        volunteers.filter(qualifs_mon1__session__in=sessions_pks)
                        .distinct()
                        .count(),
                        volunteers.filter(qualifs_actor__session__in=sessions_pks)
                        .distinct()
                        .count(),
                    ]
                )
        return dataset


class LogisticsExport(SeasonSessionsMixin):
    def get_dataset_title(self):
        return "{title} - {month_start} à {month_end} {year}".format(
            title=(gettext("Planification logistique")),
            year=self.export_year,
            month_start=season_month_start(self.export_season),
            month_end=season_month_end(self.export_season),
        )

    @property
    def export_filename(self):
        return "%s-%s-%s-%s" % (
            gettext("Logistique"),
            gettext("Mois"),
            self.export_year,
            season_verb(self.export_season),
        )

    def get_dataset(self, html=False):
        dataset = Dataset()
        dataset.headers = [
            gettext("Canton"),
            gettext("Établissement"),
            gettext("Lieu"),
            gettext("Date"),
            gettext("Heure"),
            gettext("Classes"),
            gettext("Participant·e·s"),
            gettext("Vélos loués"),
            gettext("Casques loués"),
            gettext("Logistique vélos"),
            gettext("N° de contact vélos"),
        ]
        sessions = self.get_queryset()
        sessions = sessions.annotate(
            n_qualifs=Count("qualifications"),
            n_participants=Sum("qualifications__n_participants"),
            n_bikes=Sum("qualifications__n_bikes"),
            n_helmets=Sum("qualifications__n_helmets"),
        )
        for session in sessions:
            season = session.season
            url = None
            if season and html:
                url = reverse(
                    "session-detail",
                    kwargs={"seasonpk": season.pk, "pk": session.id},
                )
            datetxt = datefilter(session.day, settings.DATE_FORMAT_COMPACT)
            timetxt = datefilter(session.begin, settings.TIME_FORMAT_SHORT)
            dataset.append(
                [
                    session.orga.address_canton,
                    session.orga.ifabbr if html else session.orga.name,
                    session.city,
                    mark_safe(linktxt.format(url=url, content=datetxt))
                    if url
                    else datetxt,
                    mark_safe(linktxt.format(url=url, content=timetxt))
                    if url
                    else timetxt,
                    session.n_qualifs,
                    session.n_participants,
                    session.n_bikes,
                    session.n_helmets,
                    session.bikes_concept,
                    session.bikes_phone,
                ]
            )
        return dataset



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.core.validators import MaxValueValidator
from django.db import models
from django.db.models import IntegerChoices, Q
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from parler.models import TranslatableModel, TranslatedFields
from sentry_sdk import capture_message as sentry_message
from sentry_sdk import configure_scope as sentry_scope
from simple_history.models import HistoricalRecords

from apps.salary.models import Timesheet
from apps.user import FORMATION_KEYS, FORMATION_M2

from .. import CHOSEN_AS_ACTOR, CHOSEN_AS_HELPER, CHOSEN_AS_LEADER
from .session import Session

# Using non-lazy translation to be exportable in XSL
CATEGORY_CHOICE_A = gettext("Agilité")
CATEGORY_CHOICE_B = gettext("Mécanique")
CATEGORY_CHOICE_C = gettext("Rencontre")

CATEGORY_CHOICES = (
    ("A", CATEGORY_CHOICE_A),
    ("B", CATEGORY_CHOICE_B),
    ("C", CATEGORY_CHOICE_C),
)


class QualificationActivity(TranslatableModel):
    translations = TranslatedFields(name=models.CharField(_("Nom"), max_length=255))
    category = models.CharField(
        _("Catégorie"), max_length=1, choices=CATEGORY_CHOICES, blank=True
    )
    history = HistoricalRecords()

    class Meta:
        verbose_name = _("Poste")
        verbose_name_plural = _("Postes")
        ordering = ["category", "pk"]

    def __str__(self):
        return self.name


def num2words(number: int) -> str:
    num_map = {
        0: _("zéro"),
        1: _("un"),
        2: _("deux"),
        3: _("trois"),
    }
    return str(num_map.get(number, str(number)))


class MonitorNumberEnum(IntegerChoices):
    """
    Represente le nombre de moniteurs 1 et 2 pour une qualif.
    - Si nombre de moniteurs = 1 alors M1 = 0, M2 = 1
    - Si nombre de moniteurs = 2 alors M1 = 1, M2 = 1
    - Si nombre de moniteurs = 3 alors M1 = 2, M2 = 1
    """

    ONE = 1, "1"
    TWO = 2, "2"
    THREE = 3, "3"

    @property
    def m1(self) -> int:
        return max(0, self.value - 1)

    @property
    def m2(self) -> int:
        return 1


class Qualification(models.Model):
    created_on = models.DateTimeField(auto_now_add=True)
    # TODO: Replace with automated or classes objects
    name = models.CharField(_("Nom de la classe"), max_length=255)
    session = models.ForeignKey(
        Session, related_name="qualifications", on_delete=models.CASCADE
    )
    class_teacher_fullname = models.CharField(
        _("Enseignant·e"), max_length=512, blank=True
    )
    class_teacher_natel = models.CharField(
        _("Natel enseignant·e"), max_length=13, blank=True
    )
    n_participants = models.PositiveSmallIntegerField(
        _("Nombre de participant·es"),
        blank=True,
        null=True,
        validators=[
            MaxValueValidator(30),
        ],
    )
    n_helpers = models.IntegerField(
        _("Nombre de moniteur·trice·s"),
        choices=MonitorNumberEnum.choices,
        default=MonitorNumberEnum.THREE,
    )

    n_bikes = models.PositiveSmallIntegerField(
        _("Nombre de vélos"),
        blank=True,
        null=True,
        validators=[
            MaxValueValidator(30),
        ],
    )
    n_helmets = models.PositiveSmallIntegerField(
        _("Nombre de casques"),
        blank=True,
        null=True,
        validators=[
            MaxValueValidator(30),
        ],
    )
    activity_A = models.ForeignKey(
        QualificationActivity,
        limit_choices_to={"category": "A"},
        verbose_name=CATEGORY_CHOICE_A,
        related_name="qualifs_A",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    activity_B = models.ForeignKey(
        QualificationActivity,
        limit_choices_to={"category": "B"},
        verbose_name=CATEGORY_CHOICE_B,
        related_name="qualifs_B",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    activity_C = models.ForeignKey(
        QualificationActivity,
        limit_choices_to={"category": "C"},
        verbose_name=CATEGORY_CHOICE_C,
        related_name="qualifs_C",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    leader = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Moniteur·trice 2"),
        related_name="qualifs_mon2",
        limit_choices_to=Q(profile__formation=FORMATION_M2),
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    helpers = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Moniteur·trice·s 1"),
        related_name="qualifs_mon1",
        limit_choices_to=Q(profile__formation__in=FORMATION_KEYS),
        blank=True,
    )
    actor = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Intervenant·e"),
        related_name="qualifs_actor",
        limit_choices_to={"profile__actor_for__isnull": False},
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    comments = models.TextField(_("Remarques"), blank=True)
    history = HistoricalRecords()

    @property
    def n_helpers_enum(self) -> MonitorNumberEnum:
        return MonitorNumberEnum(self.n_helpers)

    @cached_property
    def has_availability_incoherences(self):
        # Check les intervenants
        if (
            self.actor
            and not self.session.availability_statuses.filter(
                helper=self.actor, chosen_as=CHOSEN_AS_ACTOR
            ).exists()
        ):
            return True
        # Check les moniteurs 2
        if (
            self.leader
            and not self.session.availability_statuses.filter(
                helper=self.leader, chosen_as=CHOSEN_AS_LEADER
            ).exists()
        ):
            return True
        # Check les moniteurs 1
        for helper in self.helpers.all():
            if not self.session.availability_statuses.filter(
                helper=helper, chosen_as=CHOSEN_AS_HELPER
            ).exists():
                return True
        return False

    def fix_availability_incoherences(self):
        # many-to-many won't work without self.id.
        if not self.id:
            return

        # Check les intervenants
        if (
            self.actor
            and not self.session.availability_statuses.filter(
                helper=self.actor, chosen_as=CHOSEN_AS_ACTOR
            ).exists()
        ):
            self.actor = None
        # Check les moniteurs 2
        if (
            self.leader
            and not self.session.availability_statuses.filter(
                helper=self.leader, chosen_as=CHOSEN_AS_LEADER
            ).exists()
        ):
            self.leader = None
        # Check les moniteurs 1
        for helper in self.helpers.all():
            if not self.session.availability_statuses.filter(
                helper=helper, chosen_as=CHOSEN_AS_HELPER
            ).exists():
                self.helpers.remove(helper)

    def user_errors(self, user=None):
        errors = []
        if not self.class_teacher_fullname or not self.class_teacher_natel:
            errors.append(gettext("Enseignant·e"))
        if not self.n_participants:
            errors.append(gettext("Nombre de participant·es"))
        if self.session.orga.coordinator != user:
            if not self.leader or self.helpers.count() != self.n_helpers_enum.m1:
                errors.append(gettext("Moniteur·trice·s"))
            if not self.actor:
                errors.append(gettext("Intervenant·e"))
            if not self.activity_A or not self.activity_B or not self.activity_C:
                errors.append(gettext("Postes"))
            if self.has_availability_incoherences:
                errors.append(gettext("Incohérences de dispos"))
        if errors:
            return mark_safe(
                "<br />".join(
                    [
                        '<span class="btn-warning btn-xs disabled">'
                        '  <span class="glyphicon glyphicon-warning-sign"></span>'
                        " %s"
                        "</span>" % e
                        for e in errors
                    ]
                )
            )

    @property
    def errors(self):
        return self.user_errors(None)

    @classmethod
    def _get_label(cls, field):
        return str(cls._meta.get_field(field).verbose_name)

    @property
    def helpers_label(self):
        return self._get_label("helpers")

    @property
    def leader_label(self):
        return self._get_label("leader")

    @property
    def actor_label(self):
        return self._get_label("actor")

    def get_related_timesheets(self):
        # Do not distinct, so deletion is possible.
        # Requires distinct for display.
        return Timesheet.objects.filter(
            (
                Q(user__qualifs_actor=self)
                | Q(user__qualifs_mon2=self)
                | Q(user__qualifs_mon1=self)
            )
            & Q(date=self.session.day)
        )

    def save(self, *args, **kwargs):
        # Forcibly fix availability incoherences
        self.fix_availability_incoherences()
        with sentry_scope() as scope:
            scope.level = "info"
            scope.fingerprint = ["Qualification.save()"]
            scope.set_tag("Qualification", self)
            scope.set_tag("Qualification.id", self.id)

        sentry_message(
            "Qualification.save() : {quali}{mon2}".format(
                quali=self,
                mon2=" - Mon2: {leader} ({id})".format(
                    id=self.leader_id, leader=self.leader.get_full_name()
                )
                if self.leader
                else "",
            )
        )
        return super(Qualification, self).save(*args, **kwargs)

    class Meta:
        verbose_name = _("Qualif’")
        verbose_name_plural = _("Qualifs")
        ordering = ["session", "created_on", "name"]

    def __str__(self):
        return "{name} ({session})".format(name=self.name, session=self.session)



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models
from django.utils.translation import gettext_lazy as _

from localflavor.ch.ch_states import STATE_CHOICES


class Address(models.Model):
    address_street = models.CharField(_("Rue"), max_length=255, blank=True)
    address_no = models.CharField(_("N°"), max_length=8, blank=True)
    address_additional = models.CharField(
        _("Complément d’adresse"), max_length=255, blank=True
    )
    address_zip = models.CharField(_("NPA"), max_length=4, blank=True)
    address_city = models.CharField(_("Ville"), max_length=64, blank=True)
    address_canton = models.CharField(_("Canton"), max_length=2, blank=True)

    @property
    def address_canton_full(self):
        return [c[1] for c in STATE_CHOICES if c[0] == self.address_canton][0]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.urls import re_path

from .views import ArticleCreateView, ArticleDeleteView, ArticleUpdateView

urlpatterns = [
    re_path(r"^new/$", ArticleCreateView.as_view(), name="article-create"),
    re_path(
        r"^(?P<pk>[0-9]+)/update/$", ArticleUpdateView.as_view(), name="article-update"
    ),
    re_path(
        r"^(?P<pk>[0-9]+)/delete/$", ArticleDeleteView.as_view(), name="article-delete"
    ),
]



from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("user", "0021_remove_userprofile_office_member"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserManagedState",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        verbose_name="ID",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "canton",
                    models.CharField(
                        max_length=2,
                        verbose_name="Canton",
                        choices=[
                            ("BS", "Basel-Stadt"),
                            ("BE", "Berne"),
                            ("FR", "Fribourg"),
                            ("GE", "Geneva"),
                            ("JU", "Jura"),
                            ("LU", "Lucerne"),
                            ("NE", "Neuchatel"),
                            ("SG", "St. Gallen"),
                            ("VS", "Valais"),
                            ("VD", "Vaud"),
                            ("ZH", "Zurich"),
                        ],
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        related_name="managedstates",
                        to=settings.AUTH_USER_MODEL,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name": "Canton géré",
                "verbose_name_plural": "Cantons gérés",
            },
        ),
        migrations.AlterModelOptions(
            name="userprofile",
            options={"verbose_name": "Profil", "verbose_name_plural": "Profils"},
        ),
        migrations.AlterUniqueTogether(
            name="usermanagedstate",
            unique_together=set([("user", "canton")]),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import uuid

# Ensure Django uses the AppConfig that loads signal handlers
default_app_config = "apps.user.apps.UserConfig"

from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from localflavor.ch.ch_states import STATE_CHOICES
from memoize import memoize

STATE_CHOICES_WITH_DEFAULT = tuple(
    list(
        (
            (
                "",
                "---------",
            ),
        )
    )
    + list(STATE_CHOICES)
)

FORMATION_M1 = "M1"
FORMATION_M2 = "M2"

FORMATION_CHOICES = (
    ("", "----------"),
    (FORMATION_M1, _("Moniteur·trice 1")),
    (FORMATION_M2, _("Moniteur·trice 2")),
)
FORMATION_KEYS = [k[0] for k in FORMATION_CHOICES if k[0] != ""]


@memoize()
def formation_short(formation, real_gettext=False):
    if formation == FORMATION_M1:
        return (
            # Translators: FORMATION_M1 - Moniteur·trice 1
            gettext("M1") if real_gettext else _("M1")
        )
    elif formation == FORMATION_M2:
        return (
            # Translators: FORMATION_M2 - Moniteur·trice 2
            gettext("M2") if real_gettext else _("M2")
        )
    return ""


def get_new_username():
    return uuid.uuid4().hex[0:30]



import enum

from django.contrib.auth import get_user_model
from django.db.models import F, FloatField, IntegerField, Q, Sum
from django.db.models.functions import Cast, Upper
from django.db.models.query import QuerySet

from rolepermissions.checkers import has_permission

from apps.challenge.models.session import Session
from apps.common import DV_STATES
from apps.salary.models import Timesheet
from apps.user.models import UserProfile
from defivelo.roles import user_cantons


class TimesheetStatus(enum.IntFlag):
    TIMESHEET_MISSING = 1
    TIMESHEET_NOT_VALIDATED = 2
    TIMESHEET_VALIDATED = 4

    def __str__(self):
        return str(self.value)


User = get_user_model()


def timesheets_validation_status(year, month=None, cantons=DV_STATES):
    """
    Get timesheets validation status matrix, a {'canton': status} dict
    if month is None, calculate for the full year
    The status is:
    - True for "all timesheets validated"
    - False for "missing timesheet validations"
    - None for "No sessions in that year-month, for that canton
    """
    months = range(1, 13)
    if month:
        months = [month]

    user_cache_key = f"{'-'.join(cantons)}-{year}"
    if user_cache_key not in timesheets_validation_status.all_users:
        timesheets_validation_status.all_users[user_cache_key] = (
            UserProfile.get_users_that_worked_in_cantons(
                cantons, year=year
            ).prefetch_related("profile")
        )

    users = timesheets_validation_status.all_users[user_cache_key]

    statuses = {}

    for month_in_loop in months:
        sessions = Session.objects.filter(
            day__year=year, day__month=month_in_loop
        ).prefetch_related(
            "qualifications",
            "qualifications__leader",
            "qualifications__helpers",
            "qualifications__actor",
        )
        timesheets = get_timesheets(year=year, month=month_in_loop, users=users)

        sessions_by_user = regroup_sessions_by_user(sessions, users)
        timesheets_by_user = regroup_timesheets_by_user(timesheets)
        timesheets_statuses_by_canton = {
            canton: [
                get_timesheets_status_flags_for_user(
                    user=user,
                    sessions=sessions_by_user[user.pk],
                    timesheets=timesheets_by_user.get(user.pk, set()),
                    month_range=[month_in_loop],
                )[0]
                for user in users
                if user.profile.affiliation_canton == canton
                and user.pk in sessions_by_user
            ]
            for canton in cantons
        }
        are_all_timesheets_validated_in_month = {
            canton: all(
                [flag == TimesheetStatus.TIMESHEET_VALIDATED for flag in all_flags]
            )
            if len(all_flags)
            else None
            for canton, all_flags in timesheets_statuses_by_canton.items()
        }
        statuses[month_in_loop] = are_all_timesheets_validated_in_month

    if month:
        return statuses[month_in_loop]
    return statuses


timesheets_validation_status.all_users = {}


def get_orphaned_timesheets_per_month(year, users, month=None, cantons=DV_STATES):
    """
    Get orphaned timesheets validation status matrix, a {'canton': status} dict
    if month is None, calculate for the full year
    """
    months = range(1, 13)
    if month:
        months = [month]

    if not cantons:
        cantons = DV_STATES

    orphaned_timesheets_year = {}

    for month_in_loop in months:
        sessions = Session.objects.filter(
            day__year=year, day__month=month_in_loop
        ).prefetch_related(
            "qualifications",
            "qualifications__leader",
            "qualifications__helpers",
            "qualifications__actor",
        )
        timesheets = get_timesheets(year=year, month=month_in_loop, users=users)

        sessions_by_user = regroup_sessions_by_user(sessions, users)
        timesheets_by_user = regroup_timesheets_by_user(timesheets)
        orphaned_timesheets_month = set()
        for user in users:
            try:
                if user.profile.affiliation_canton in cantons:
                    orphaned_timesheets_month.update(
                        get_timesheets_without_corresponding_session(
                            sessions_by_user.get(user.pk, set()),
                            timesheets_by_user.get(user.pk, set()),
                        )
                    )
            except User.profile.RelatedObjectDoesNotExist:
                pass

        orphaned_timesheets_year[month_in_loop] = orphaned_timesheets_month

    if month:
        return orphaned_timesheets_year[month_in_loop]
    return orphaned_timesheets_year


def get_timesheets_status_matrix(year, users, cantons: list | None = None):
    """
    Return a dict of `{user: [TimesheetStatus]}` for all users who worked during the
    given `year`. Only users who `user` has access to (either because the user can see
    all cantons, or because they're affiliated in the canton the user manages) are
    returned.
    The cantons filter will apply on the Sessions & timesheets, not on the users.
    """
    sessions = (
        Session.objects.filter(day__year=year).annotate(
            orga_canton=Upper(F("orga__address_canton"))
        )
    ).prefetch_related("qualifications", "qualifications__helpers")
    timesheets = get_timesheets(year=year, users=users)
    if cantons is not None:
        sessions = sessions.filter(orga_canton__in=[c.upper() for c in cantons])
        timesheets = timesheets.filter(date__in=sessions.values_list("day"))

    sessions_by_user = regroup_sessions_by_user(sessions, users)
    timesheets_by_user = regroup_timesheets_by_user(timesheets)

    timesheets_status_matrix = {
        user: get_timesheets_status_flags_for_user(
            user, sessions_by_user[user.pk], timesheets_by_user.get(user.pk, set())
        )
        for user in users
        if user.pk in sessions_by_user
    }

    return timesheets_status_matrix


def get_timesheets_status_flags_for_user(
    user, sessions, timesheets, month_range=range(1, 13)
):
    """
    Return a list of `TimesheetStatus` for the given `user`, `sessions` and
    `timesheets`. The list always contains 12 elements, one for each month (or for a selection of months)
    """
    months_with_missing_timesheets = get_months_without_timesheets(sessions, timesheets)
    timesheets_by_month = regroup_timesheets_by_month(timesheets)

    def _get_flags(month):
        flags = 0

        if month in months_with_missing_timesheets:
            flags |= TimesheetStatus.TIMESHEET_MISSING

        for timesheet in timesheets_by_month.get(month, []):
            if timesheet.validated_at:
                flags |= TimesheetStatus.TIMESHEET_VALIDATED
            elif not timesheet.validated_at:
                flags |= TimesheetStatus.TIMESHEET_NOT_VALIDATED

        return flags

    return [_get_flags(month) for month in month_range]


def regroup_sessions_by_user(sessions, users):
    """
    Regroup the given `Session` queryset by user (only keeping users in `users`) and
    return a dict `{user_id: {session1, session2, ...}}`.
    """
    sessions_by_user = {}
    users_ids = {user.pk for user in users}

    for session in sessions:
        for qualification in session.qualifications.all():
            if qualification.leader_id and qualification.leader_id in users_ids:
                sessions_by_user.setdefault(qualification.leader_id, set()).add(session)

            if qualification.actor_id and qualification.actor_id in users_ids:
                sessions_by_user.setdefault(qualification.actor_id, set()).add(session)

            for helper in qualification.helpers.all():
                if helper.pk in users_ids:
                    sessions_by_user.setdefault(helper.pk, set()).add(session)

    return sessions_by_user


def regroup_timesheets_by_user(timesheets):
    """
    Regroup the given `Timesheet` queryset by user and return a dict `{user_id:
    {timesheet1, timesheet2, ...}}`.
    """
    timesheets_by_user = {}

    for timesheet in timesheets:
        timesheets_by_user.setdefault(timesheet.user_id, set()).add(timesheet)

    return timesheets_by_user


def regroup_timesheets_by_month(timesheets):
    """
    Regroup the given `Timesheet` queryset by month and return a dict of `{month:
    [timesheets]}`, where `month` is a month number (1 - 12) and `timesheets` the list
    of associated timesheets.
    """
    timesheets_by_month = {}
    for timesheet in timesheets:
        timesheets_by_month.setdefault(timesheet.date.month, []).append(timesheet)

    return timesheets_by_month


def get_months_without_timesheets(sessions, timesheets):
    """
    Return a set of month numbers (1 - 12) that have sessions that don't have any
    associated timesheet.
    """
    days_with_missing_timesheets = set(session.day for session in sessions) - set(
        timesheet.date for timesheet in timesheets
    )

    return {date.month for date in days_with_missing_timesheets}


def get_timesheets_without_corresponding_session(sessions, timesheets):
    """
    Return a set of timesheets without a session in the corresponding day
    """
    session_days = [session.day for session in sessions]
    return set(
        timesheet for timesheet in timesheets if timesheet.date not in session_days
    )


def get_timesheets(year, users, month=None):
    """
    Return the timesheets of the given `users` for the given `year` (and eventually `month`).
    """
    kwargs = {"date__year": year, "user__in": users}
    if month:
        kwargs["date__month"] = month
    return Timesheet.objects.filter(**kwargs)


def get_timesheets_amount_by_month(year, users, cantons: list | None = None):
    """
    Return the total amount of timesheets for the given year and the given users, as a
    list of 12 elements, 0 being January and 11 being December.
    """
    timesheets = get_timesheets(year=year, users=users)

    if cantons is not None:
        sessions = (
            Session.objects.filter(day__year=year)
            .annotate(orga_canton=Upper(F("orga__address_canton")))
            .filter(orga_canton__in=[c.upper() for c in cantons])
        )
        timesheets = timesheets.filter(date__in=sessions.values_list("day"))

    timesheets_by_month = regroup_timesheets_by_month(timesheets)

    total_by_month = [
        sum(
            timesheet.get_total_amount() for timesheet in timesheets_by_month.get(i, [])
        )
        for i in range(1, 13)
    ]

    return total_by_month


def get_visible_users(user):
    """
    Return a queryset of `User` objects the given `user` has access to.
    """

    if has_permission(user, "cantons_all"):
        qs = User.objects.all()
    elif has_permission(user, "cantons_mine"):
        qs = UserProfile.get_users_that_worked_in_cantons(
            user_cantons(user), [Q(pk=user.pk)]
        )
    else:
        qs = User.objects.filter(pk=user.pk)

    return qs.prefetch_related("profile")


def get_users_with_missing_timesheets(year: int, month: int, users):
    return [
        user
        for user, flags in get_timesheets_status_matrix(year, users).items()
        if flags[month - 1] & TimesheetStatus.TIMESHEET_MISSING
    ]


def get_missing_timesheet_status_per_month(timesheets_status_matrix):
    """
    Return a list of booleans (one per month) that indicate if any user has missing
    timesheets for each month.
    """
    return [
        any(
            statuses[month_index] & TimesheetStatus.TIMESHEET_MISSING
            for _, statuses in timesheets_status_matrix.items()
        )
        for month_index in range(0, 12)
    ]


def get_salary_details_list(object_list: QuerySet) -> QuerySet:
    """
    Return the queryset of salary details needed in multiple exports
    """

    # Django queries to convert the ignore Bool (0 if not ignored, 1 if ignored)
    # into an included Bool (1 if taken into account, 0 if ignored)
    included = 1 - Cast(F("ignore"), IntegerField())
    included_float = Cast(included, FloatField())

    return object_list.values("user").annotate(
        employee_code=F("user__profile__employee_code"),
        last_name=F("user__last_name"),
        first_name=F("user__first_name"),
        time_helper=Sum(F("time_helper") * included_float),
        actor_count=Sum(F("actor_count") * included),
        leader_count=Sum(F("leader_count") * included),
        overtime=Sum(F("overtime") * included_float),
        traveltime=Sum(F("traveltime") * included_float),
    )



# Generated by Django 2.2.12 on 2020-05-18 01:22

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("orga", "0007_organization_organizator_migrate"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="organization",
            name="coordinator_email",
        ),
        migrations.RemoveField(
            model_name="organization",
            name="coordinator_fullname",
        ),
        migrations.RemoveField(
            model_name="organization",
            name="coordinator_natel",
        ),
        migrations.RemoveField(
            model_name="organization",
            name="coordinator_phone",
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.utils.translation import gettext_lazy as _

from memoize import memoize
from rolepermissions.checkers import has_permission as uncached_has_permission
from rolepermissions.roles import AbstractUserRole

from apps.common import DV_STATES


# Override 'has_permission' from rolepermissions to add memoization for performance reasons
@memoize()
def has_permission(user, permission_name):
    return uncached_has_permission(user, permission_name)


@memoize()
def user_cantons(user):
    """
    List of the cantons _managed_ by this user
    """
    if has_permission(user, "cantons_all"):
        return DV_STATES
    elif has_permission(user, "cantons_mine"):
        return [m.canton for m in user.managedstates.all()]
    else:
        raise LookupError("No user cantons")


class Collaborator(AbstractUserRole):
    """
    Moniteur·trice 1, 2 ou Intervenant
    """

    available_permissions = {
        "challenge_see_all_orga": True,
        "user_view_list": True,
        "user_view_list_non_collaborator": False,
        "user_export_all_fields": False,
        "timesheet": True,
    }


class StateManager(AbstractUserRole):
    """
    Chargé de projet, responsable d’un ou plusieurs cantons
    """

    available_permissions = {
        "cantons_all": False,
        "cantons_mine": True,
        "user_view_list": True,
        "user_view_list_non_collaborator": True,
        "user_export_all_fields": True,
        "user_detail_other": True,
        "user_edit_other": True,
        "user_crud_dv_public_fields": True,
        "user_crud_dv_private_fields": True,
        "user_can_send_credentials": True,
        "user_create": True,
        "orga_detail_all": True,
        "orga_crud": True,
        "orga_edit_address": True,
        "orga_show": True,
        "orga_edit": True,
        "challenge_season_crud": True,
        "challenge_see_all_orga": True,
        "challenge_session_crud": True,
        "challenge_invoice_cru": True,
        "challenge_invoice_reset_to_draft": False,
        "challenge_season_see_state_planning": True,
        "settings_crud": True,
        "timesheet": True,
        "timesheet_editor": True,
        "registration_validate": True,
        "assign_only_coordinator_role": True,
    }


class PowerUser(AbstractUserRole):
    """
    Bureau
    """

    available_permissions = {
        "cantons_all": True,
        "cantons_mine": True,
        "user_view_list": True,
        "user_view_list_non_collaborator": True,
        "user_export_all_fields": True,
        "user_detail_other": True,
        "user_edit_other": True,
        "user_create": True,
        "user_crud_dv_public_fields": True,
        "user_crud_dv_private_fields": True,
        "user_edit_employee_code": True,
        "user_can_send_credentials": True,
        "user_can_resend_credentials": True,
        "user_deletions": True,
        "user_set_role": True,
        "user_mark_inactive": True,
        "home_article_cud": True,
        "orga_detail_all": True,
        "orga_crud": True,
        "orga_edit_address": True,
        "orga_show": True,
        "orga_edit": True,
        "challenge_invoice_cru": True,
        "challenge_invoice_reset_to_draft": True,
        "challenge_see_all_orga": True,
        "challenge_session_crud": True,
        "challenge_season_crud": True,
        "challenge_season_see_state_planning": True,
        "settings_crud": True,
        "timesheet": True,
        "timesheet_control": True,
        "timesheet_editor": True,
        "export_yearly_invoices": True,
    }


class Coordinator(AbstractUserRole):
    """
    Coordinateur d’établissement
    """

    available_permissions = {
        "coordinator_home": True,
        "orga_show": True,
        "orga_edit": True,
        "challenge_see_all_orga": False,
        "registration_create": True,
    }


DV_AVAILABLE_ROLES = (
    (None, _("Aucun rôle")),
    ("collaborator", _("Collabora·teur·trice")),
    ("state_manager", _("Chargé·e de projet")),
    ("coordinator", _("Coordina·teur·trice")),
    ("power_user", _("Bureau de coordination")),
)

DV_AUTOMATIC_ROLES = ["collaborator"]



# Generated by Django 2.2.9 on 2020-04-08 12:41

from django.db import migrations


def session_to_historical(apps, schema_editor):
    # We can't import the InvoiceLine model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    InvoiceLine = apps.get_model("challenge", "InvoiceLine")
    HistoricalSession = apps.get_model("challenge", "HistoricalSession")
    for il in InvoiceLine.objects.all():
        hs = (
            HistoricalSession.objects.filter(id=il.session_id)
            .order_by("-history_id")
            .last()
        )
        il.historical_session = hs
        il.save()


def historical_to_session(apps, schema_editor):
    # Don't care, no info lost.
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0061_invoiceline_historical_session"),
    ]

    operations = [
        migrations.RunPython(session_to_historical, historical_to_session),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.forms import modelform_factory
from django.utils.translation import gettext_lazy as _

from django_filters import DateFilter
from localflavor.ch.forms import (
    CHSocialSecurityNumberField,
    CHStateSelect,
    CHZipCodeField,
)
from localflavor.generic import forms as localforms
from localflavor.generic.countries.sepa import IBAN_SEPA_COUNTRIES
from rolepermissions.roles import get_user_roles

from apps.common import DV_STATE_CHOICES
from apps.common.forms import (
    BS3CountriesField,
    CHPhoneNumberField,
    SelectWithDisabledValues,
    SwissDateField,
)
from apps.orga.models import Organization
from defivelo.roles import DV_AUTOMATIC_ROLES, DV_AVAILABLE_ROLES

from ..common.fields import CheckboxMultipleChoiceField
from . import STATE_CHOICES_WITH_DEFAULT
from .models import UserProfile


class SwissDateFilter(DateFilter):
    field_class = SwissDateField


class SimpleUserProfileForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        allow_email = kwargs.pop("allow_email", False)
        kwargs.pop("affiliation_canton_required", None)
        kwargs.pop("cantons", None)
        profile_fields = ["natel", "phone", "language", "comments"]
        if kwargs.pop("affiliation_canton", False):
            profile_fields.append("activity_cantons")
            profile_fields.append("affiliation_canton")
        super().__init__(*args, **kwargs)
        self.fields.update(
            modelform_factory(UserProfile, fields=profile_fields)().fields
        )

        # Some manual fixes lost by importing UserProfile
        for fieldname, fieldclass in {
            "natel": CHPhoneNumberField(required=False),
            "phone": CHPhoneNumberField(required=False),
        }.items():
            label = self.fields[fieldname].label
            self.fields[fieldname] = fieldclass
            self.fields[fieldname].label = label

        if not allow_email and "email" in self.fields:
            del self.fields["email"]

        for field in (
            "first_name",
            "last_name",
            "email",
            "language",
        ):
            if field in self.fields:
                self.fields[field].required = True

    class Meta:
        model = get_user_model()
        fields = [
            "first_name",
            "last_name",
            "email",
        ]

    def clean_email(self):
        # Ideally, this should get checked by the User model
        email = self.cleaned_data["email"]
        existing_users = self._meta.model.objects.filter(email=email)
        if self.instance and self.instance.pk is not None:
            existing_users = existing_users.exclude(pk=self.instance.pk)
        if existing_users.exists():
            raise ValidationError(
                _("Un·e utilisateur·rice avec cette adresse e-mail existe déjà.")
            )
        return email

    def clean(self):
        """
        Vérifie les contraintes croisées
        """
        cleaned_data = super().clean()

        try:
            affiliation_canton = cleaned_data["affiliation_canton"]
        except KeyError:
            affiliation_canton = None
        actor_for = (
            cleaned_data["actor_for"]
            if "actor_for" in cleaned_data
            else self.instance.profile.actor_for.all()
        )
        formation = (
            cleaned_data["formation"]
            if "formation" in cleaned_data
            else self.instance.profile.formation
        )
        if not affiliation_canton and (actor_for or formation):
            self.add_error(
                "affiliation_canton",
                ValidationError(
                    _(
                        "Les moniteur·trice·s / intervenant·e·s ont besoin d’un canton d’affiliation."
                    )
                ),
            )
        return cleaned_data


class UserProfileForm(SimpleUserProfileForm):
    def __init__(self, *args, **kwargs):
        # Whether to permit no-affiliation-canton creation
        affiliation_canton_required = kwargs.pop("affiliation_canton_required", True)
        cantons = kwargs.pop("cantons", None)
        super().__init__(*args, **kwargs)
        # Import all generated fields from UserProfile
        self.fields.update(modelform_factory(UserProfile, exclude=("user",))().fields)
        # Some manual fixes lost by importing UserProfile
        for fieldname, fieldclass in {
            "address_zip": CHZipCodeField(required=False),
            "birthdate": SwissDateField(required=False),
            "natel": CHPhoneNumberField(required=False),
            "phone": CHPhoneNumberField(required=False),
            "nationality": BS3CountriesField(required=False),
            "social_security": CHSocialSecurityNumberField(required=False),
            "address_canton": forms.ChoiceField(
                widget=CHStateSelect,
                choices=STATE_CHOICES_WITH_DEFAULT,
                required=False,
            ),
            "iban": localforms.IBANFormField(
                include_countries=IBAN_SEPA_COUNTRIES,
                required=False,
            ),
        }.items():
            label = self.fields[fieldname].label
            self.fields[fieldname] = fieldclass
            self.fields[fieldname].label = label
        # Some precisions
        self.fields["work_permit"].widget.attrs["placeholder"] = _("… si pas suisse")
        self.fields["tax_jurisdiction"].widget.attrs["placeholder"] = _(
            "… si pas en Suisse"
        )

        for field in (
            "first_name",
            "last_name",
            "email",
            "language",
            "affiliation_canton",
            "status",
        ):
            if field in self.fields:
                self.fields[field].required = True

        if cantons and "affiliation_canton" in self.fields:
            # if we edit a user from a other canton we cannot move him
            if (
                "affiliation_canton" in self.initial
                and self.initial["affiliation_canton"]
                and self.initial["affiliation_canton"] not in cantons
            ):
                self.fields["affiliation_canton"].disabled = True
            # Only permit edition within the allowed cantons
            choices = self.fields["affiliation_canton"].choices
            choices = (
                (k, v)
                for (k, v) in choices
                if k in cantons
                or k == ""
                or k == self.initial.get("affiliation_canton", "")
            )
            self.fields["affiliation_canton"].required = affiliation_canton_required
            self.fields["affiliation_canton"].choices = choices


class UserAssignRoleForm(forms.Form):
    role = forms.ChoiceField(
        label=_("Niveau d’accès"),
        choices=DV_AVAILABLE_ROLES,
        required=False,
        widget=SelectWithDisabledValues(
            choices=DV_AVAILABLE_ROLES,
            disabled_values=DV_AUTOMATIC_ROLES,
        ),
    )
    managed_states = CheckboxMultipleChoiceField(
        label=_("Cantons gérés"), choices=sorted(DV_STATE_CHOICES), required=False
    )
    managed_organizations = CheckboxMultipleChoiceField(
        label=_("Établissements gérés"),
        required=False,
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super().__init__(*args, **kwargs)
        roles = get_user_roles(user)
        self.fields["managed_organizations"].choices = [
            (orga.id, orga.name)
            for orga in Organization.objects.filter(
                Q(coordinator__isnull=True) | Q(coordinator=user)
            ).all()
        ]
        self.initial = {
            "role": roles[0].get_name() if len(roles) >= 1 else None,
            "managed_states": list(
                user.managedstates.all().values_list("canton", flat=True)
            ),
            "managed_organizations": list(
                user.managed_organizations.all().values_list("id", flat=True)
            ),
        }

    def save(self):
        role = self.cleaned_data["role"]
        self.user.profile.set_role(role)

        managed_states = self.cleaned_data["managed_states"]
        if role != "state_manager":
            managed_states = []
        self.user.profile.set_statemanager_for(managed_states)

        managed_organizations = self.cleaned_data["managed_organizations"]
        if role != "coordinator":
            managed_organizations = []
        self.user.managed_organizations.set(
            Organization.objects.filter(pk__in=managed_organizations)
        )



# Generated by Django 1.11.4 on 2017-08-14 13:52
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0033_auto_20170131_2000"),
    ]

    operations = [
        migrations.RenameField(
            model_name="session",
            old_name="organization",
            new_name="orga",
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0003_auto_20150901_1742"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="formation",
            field=models.CharField(
                verbose_name="Formation",
                max_length=2,
                choices=[
                    ("", "----------"),
                    ("M1", "Moniteur 1"),
                    ("M2", "Moniteur 2"),
                ],
                blank=True,
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0001_initial"),
        ("challenge", "0016_auto_20150911_1738"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="session",
            name="id",
        ),
        migrations.AddField(
            model_name="session",
            name="address_ptr",
            field=models.OneToOneField(
                to="common.Address",
                parent_link=True,
                auto_created=True,
                serialize=False,
                default=1,
                primary_key=True,
                on_delete=models.CASCADE,
            ),
            preserve_default=False,
        ),
    ]



from django.core import validators
from django.db import migrations, models

from apps.common import DV_SEASON_AUTUMN, DV_SEASON_LAST_SPRING_MONTH, DV_SEASON_SPRING


def season_to_months(apps, schema_editor):
    Season = apps.get_model("challenge", "Season")
    for s in Season.objects.all():
        if s.season == DV_SEASON_SPRING:
            s.month_start = 1
            s.n_months = DV_SEASON_LAST_SPRING_MONTH
        if s.season == DV_SEASON_AUTUMN:
            s.month_start = DV_SEASON_LAST_SPRING_MONTH + 1
            s.n_months = 12 - DV_SEASON_LAST_SPRING_MONTH
        s.save()


def months_to_season(apps, schema_editor):
    Season = apps.get_model("challenge", "Season")
    for s in Season.objects.all():
        # We would loose data by applying this reverse migration.
        s.season = DV_SEASON_AUTUMN
        if s.month_start <= DV_SEASON_LAST_SPRING_MONTH:
            s.season = DV_SEASON_SPRING
        s.save()


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0069_use_u2019_quote_qualif"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalseason",
            name="month_start",
            field=models.PositiveSmallIntegerField(
                choices=[
                    (1, "January"),
                    (2, "February"),
                    (3, "March"),
                    (4, "April"),
                    (5, "May"),
                    (6, "June"),
                    (7, "July"),
                    (8, "August"),
                    (9, "September"),
                    (10, "October"),
                    (11, "November"),
                    (12, "December"),
                ],
                null=True,
                verbose_name="Mois de début",
            ),
        ),
        migrations.AddField(
            model_name="historicalseason",
            name="n_months",
            field=models.PositiveSmallIntegerField(
                default=1,
                validators=[
                    validators.MinValueValidator(1),
                    validators.MaxValueValidator(24),
                ],
                verbose_name="Nombre de mois",
            ),
        ),
        migrations.AddField(
            model_name="season",
            name="month_start",
            field=models.PositiveSmallIntegerField(
                choices=[
                    (1, "January"),
                    (2, "February"),
                    (3, "March"),
                    (4, "April"),
                    (5, "May"),
                    (6, "June"),
                    (7, "July"),
                    (8, "August"),
                    (9, "September"),
                    (10, "October"),
                    (11, "November"),
                    (12, "December"),
                ],
                null=True,
                verbose_name="Mois de début",
            ),
        ),
        migrations.AddField(
            model_name="season",
            name="n_months",
            field=models.PositiveSmallIntegerField(
                default=1,
                validators=[
                    validators.MinValueValidator(1),
                    validators.MaxValueValidator(24),
                ],
                verbose_name="Nombre de mois",
            ),
        ),
        migrations.RunPython(season_to_months, months_to_season),
    ]



from django.core import mail
from django.test import TestCase
from django.urls import reverse

from apps.user import FORMATION_M1
from apps.user.tests.factories import UserFactory
from defivelo.tests.utils import StateManagerAuthClient

from ..models import HelperSessionAvailability
from .factories import QualificationFactory, SeasonFactory, SessionFactory


class SeasonReminderTests(TestCase):
    def setUp(self):
        # State manager client so we can access season management views
        self.client = StateManagerAuthClient()

        # Create a season in one of the manager's cantons
        self.season = SeasonFactory()
        mycantons = [c.canton for c in self.client.user.managedstates.all()]
        self.season.cantons = mycantons or self.season.cantons
        self.season.save()

        # Ensure there are at least two sessions with qualifications in-season
        self.session = SessionFactory(orga__address_canton=self.season.cantons[0])
        # Put the session within the season dates
        self.session.day = self.season.begin
        self.session.save()
        # Attach at least one qualification so it appears in sessions_with_qualifs
        QualificationFactory(session=self.session)

        self.session2 = SessionFactory(orga__address_canton=self.season.cantons[0])
        self.session2.day = self.season.begin
        self.session2.save()
        QualificationFactory(session=self.session2)

        # Two users in the season canton
        self.user_no_avail = UserFactory(
            profile__affiliation_canton=self.season.cantons[0],
            profile__formation=FORMATION_M1,
        )
        self.user_with_avail = UserFactory(
            profile__affiliation_canton=self.season.cantons[0],
            profile__formation=FORMATION_M1,
        )

        # Give one availability to user_with_avail in this season
        HelperSessionAvailability.objects.update_or_create(
            session=self.session,
            helper=self.user_with_avail,
            defaults={"availability": "y"},
        )

    def test_recipients_excludes_users_with_any_availability(self):
        url = reverse("season-availability-reminder", kwargs={"pk": self.season.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        recipients = list(response.context["recipients"])  # queryset evaluated in view
        # user_no_avail has none -> included; user_with_avail has only partial -> included too
        self.assertIn(self.user_no_avail, recipients)
        self.assertIn(self.user_with_avail, recipients)

    def test_post_sends_emails_and_sets_timestamp(self):
        url = reverse("season-availability-reminder", kwargs={"pk": self.season.pk})

        # Pre-condition: no timestamp
        self.assertIsNone(self.season.availability_reminder_sent_at)

        # Send the emails
        response = self.client.post(url, {"sendemail": "on"}, follow=True)
        self.assertIn(response.status_code, [200, 302])

        # Expect two emails sent (users without full availability across all sessions)
        self.assertEqual(len(mail.outbox), 2)
        to_addrs = {mail.outbox[0].to[0], mail.outbox[1].to[0]}
        self.assertTrue(any(self.user_no_avail.email in r for r in to_addrs))
        self.assertTrue(any(self.user_with_avail.email in r for r in to_addrs))

        # Timestamp should be set on season
        self.season.refresh_from_db()
        self.assertIsNotNone(self.season.availability_reminder_sent_at)



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.forms import Form as DjangoEmptyForm
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import FormView

from rolepermissions.mixins import HasPermissionsMixin

from .. import FORMATION_KEYS
from ..models import (
    USERSTATUS_ACTIVE,
    USERSTATUS_INACTIVE,
    USERSTATUS_RESERVE,
    UserProfile,
)


class MarkInactive(HasPermissionsMixin, FormView):
    template_name = "actions/mark_inactive.html"
    success_message = _("Tous les moniteurs ont été marqués inactifs.")
    success_url = reverse_lazy("user-list")
    form_class = DjangoEmptyForm
    required_permission = "user_mark_inactive"

    def get_affected_accounts(self):
        # Rend inactifs
        return (
            UserProfile.objects
            # … tous les moniteurs actifs ou réserve
            .filter(
                formation__in=FORMATION_KEYS,
                status__in=[
                    USERSTATUS_ACTIVE,
                    USERSTATUS_RESERVE,
                ],
            )
            # … qui n’ont pas de rôles
            .exclude(user__groups__isnull=False)
            .order_by("user__first_name", "user__last_name")
            .prefetch_related("user")
        )

    def get_context_data(self, **kwargs):
        context = super(MarkInactive, self).get_context_data(**kwargs)
        context["affected_accounts"] = self.get_affected_accounts()
        return context

    def form_valid(self, form):
        # Proceed
        self.get_affected_accounts().update(
            status=USERSTATUS_INACTIVE, status_updatetime=timezone.now()
        )
        return super(MarkInactive, self).form_valid(form)



# Generated by Django 2.2.16 on 2020-11-02 17:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0062_add_schwyz"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userprofile",
            name="affiliation_canton",
            field=models.CharField(
                blank=True,
                choices=[
                    ("", "---------"),
                    ("AG", "Aargau"),
                    ("AR", "Appenzell Ausserrhoden"),
                    ("BS", "Basel-Stadt"),
                    ("BL", "Basel-Land"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("GR", "Graubuenden"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SZ", "Schwyz"),
                    ("SO", "Solothurn"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                    ("WS", "Haut-Valais"),
                ],
                max_length=5,
                verbose_name="Canton d’affiliation",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="cresus_employee_number",
            field=models.CharField(
                blank=True, max_length=63, verbose_name="Numéro d’employé Crésus"
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="tax_jurisdiction",
            field=models.CharField(
                blank=True, max_length=511, verbose_name="Lieu d’imposition"
            ),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015, 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from . import get_env_variable
from .base import *

DEBUG = bool(get_env_variable("DEBUG", True))
TEMPLATES[0]["OPTIONS"]["debug"] = DEBUG
SECRET_KEY = "notsosecret"
NEVERCACHE_KEY = "notsosecret"

COMPRESS_ENABLED = False

INSTALLED_APPS += (
    "debug_toolbar",
    "django_extensions",
)


def show_toolbar(request):
    from django.conf import settings

    return settings.DEBUG


DEBUG_TOOLBAR_CONFIG = {
    "SHOW_TOOLBAR_CALLBACK": show_toolbar,
    "INTERCEPT_REDIRECTS": False,
}

INTERNAL_IPS = (
    "127.0.0.1",
    "10.0.3.1",
)
MIDDLEWARE += [
    "debug_toolbar.middleware.DebugToolbarMiddleware",
]

# Adapt Stronghold for Debug
STRONGHOLD_PUBLIC_URLS += [r"^/__debug__/.*$"]



# Generated by Django 4.2.6 on 2025-11-13 14:18

from django.db import migrations
translation_mapping = {
    "de": [
        ("Veloreisender", "Veloreisende·r"),
        ("Velokurier", "Velokurier·in"),
        ("Mechaniker", "Mechaniker·in"),
        ("Polizist", "Polizist·in"),
        ("Stadtplaner", "Stadtplaner·in"),
    ],
    "fr": [
        ("Coursier/Livreur", "Coursier·ère/Livreur·euse"),
        ("Cyclovoyageur", "Cyclovoyageur·euse"),
        ("Mécanicien", "Mécanicien·ne"),
        ("Policier", "Policier·ère"),
        (" Autre", "Autre"), # Fix spacing
    ]
}



def apply_inclusive_updates(apps, schema_editor):
    Translation = apps.get_model("challenge", "QualificationActivityTranslation")

    for lang_code, pairs in translation_mapping.items():
        for old, new in pairs:
            Translation.objects.filter(
                language_code=lang_code,
                name=old
            ).update(name=new)
    refresh_cache()


def reverse_inclusive_updates(apps, schema_editor):
    Translation = apps.get_model("challenge", "QualificationActivityTranslation")

    for lang_code, pairs in translation_mapping.items():
        for old, new in pairs:
            Translation.objects.filter(
                language_code=lang_code,
                name=new
            ).update(name=old)
    refresh_cache()


def refresh_cache():
    from django.core.cache import cache
    cache.clear()


class Migration(migrations.Migration):

    dependencies = [
        ('challenge', '0086_inclusive_language'),
    ]

    operations = [
        migrations.RunPython(
            apply_inclusive_updates,
            reverse_inclusive_updates
        ),
    ]



# Generated by Django 2.2.9 on 2020-03-05 16:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0057_invoice_invoiceline"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="status",
            field=models.IntegerField(
                choices=[(0, "Brouillon"), (1, "Validée")],
                default=0,
                verbose_name="Statut",
            ),
        ),
    ]



# Generated by Django 1.11.4 on 2017-08-21 15:02
from __future__ import unicode_literals

import django.db.models.manager
import django_countries.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0033_auto_20170818_1626"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="userprofile",
            managers=[
                ("objects_existing", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name="userprofile",
            name="nationality",
            field=django_countries.fields.CountryField(
                default="CH", max_length=2, verbose_name="Nationalité"
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="work_permit",
            field=models.CharField(
                blank=True,
                max_length=255,
                verbose_name="Permis de travail (si pas suisse)",
            ),
        ),
    ]



from datetime import datetime

from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from apps.challenge.forms.registration import (
    OrganizationSelectionForm,
    RegistrationConfirmForm,
    RegistrationFormSet,
    RegistrationValidationFormSet,
)
from apps.challenge.models.registration import REGISTRATION_DAY_TIMES, Registration
from apps.orga.models import Organization
from defivelo.roles import has_permission, user_cantons
from defivelo.templatetags.dv_filters import dv_season_url


def register(request):
    if not request.user.managed_organizations.exists():
        raise PermissionDenied

    if request.method == "POST":
        organization_form = OrganizationSelectionForm(
            request.POST, coordinator=request.user
        )
        if organization_form.is_valid():
            formset = RegistrationFormSet(
                request.POST,
                form_kwargs={
                    "coordinator": request.user,
                    "organization": organization_form.cleaned_data["organization"],
                },
            )

            if formset.is_valid():
                request.session["new_registration"] = {
                    "organization": organization_form.cleaned_data["organization"].id,
                    "lines": formset.serialize(),
                }

                return HttpResponseRedirect(reverse("registration-confirm"))
        else:
            formset = RegistrationFormSet(form_kwargs={"coordinator": request.user})
    else:
        submitted = request.session.get("new_registration")
        initial = []
        if submitted:
            for line in submitted["lines"]:
                initial.append(
                    {
                        "date": datetime.fromisoformat(line["date"]),
                        "day_time": line["day_time"],
                        "classes_amount": line["classes_amount"],
                    }
                )

        organization_form = OrganizationSelectionForm(coordinator=request.user)
        formset = RegistrationFormSet(
            form_kwargs={"coordinator": request.user}, initial=initial
        )

    return TemplateResponse(
        request=request,
        context={"form": organization_form, "formset": formset},
        template="challenge/registration.html",
    )


def register_confirm(request):
    if not request.user.managed_organizations.exists():
        raise PermissionDenied

    data = request.session.get("new_registration")
    if not data:
        raise SuspiciousOperation(_("Aucune inscription à valider"))

    organization = get_object_or_404(Organization, pk=data["organization"])
    lines = []
    total_classes = 0
    for line in data["lines"]:
        lines.append(
            {
                "date": datetime.fromisoformat(line["date"]),
                "day_time": dict(REGISTRATION_DAY_TIMES)[line["day_time"]],
                "classes_amount": line["classes_amount"],
            }
        )
        total_classes += int(line["classes_amount"])

    if request.method == "POST":
        form = RegistrationConfirmForm(request.POST)
        if form.is_valid():
            for line in data["lines"]:
                Registration.objects.create(
                    date=datetime.fromisoformat(line["date"]),
                    day_time=line["day_time"],
                    classes_amount=line["classes_amount"],
                    coordinator=request.user,
                    organization=organization,
                )
            del request.session["new_registration"]
            organization.notify_new_registrations(request)
            return redirect(reverse("home"))
    else:
        form = RegistrationConfirmForm()

    return TemplateResponse(
        request=request,
        context={
            "organization": organization,
            "lines": lines,
            "total_classes": total_classes,
            "form": form,
        },
        template="challenge/registration_confirm.html",
    )


def register_validate(request):
    if not has_permission(request.user, "registration_validate"):
        raise PermissionDenied

    organizations = Organization.objects.filter(
        registration__isnull=False,
        registration__is_archived=False,
        address_canton__in=user_cantons(request.user),
    ).distinct()

    if request.method == "POST":
        organization = get_object_or_404(
            Organization, pk=request.POST.get("form-organization-id")
        )
        formset = RegistrationValidationFormSet(
            organization=organization,
            data=request.POST,
        )

        if formset.is_valid():
            formset.save()
            organization.notify_registrations_validated(request)
            return redirect(dv_season_url())

        else:
            data = []
            for other_organization in organizations:
                if organization == other_organization:
                    # This organization's formset has errors
                    data.append((organization, formset))
                else:
                    # These organizations' formsets haven't been submitted
                    data.append(
                        (
                            other_organization,
                            RegistrationValidationFormSet(
                                organization=other_organization,
                            ),
                        )
                    )
    else:
        data = [
            (
                organization,
                RegistrationValidationFormSet(
                    organization=organization,
                ),
            )
            for organization in organizations
        ]

    return TemplateResponse(
        request=request,
        context={"organizations": data},
        template="challenge/registration_validate.html",
    )



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date, timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils.dates import MONTHS
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy as _p

from simple_history.models import HistoricalRecords

from apps.common import (
    DV_SEASON_AUTUMN,
    DV_SEASON_CHOICES,
    DV_SEASON_LAST_SPRING_MONTH,
    DV_SEASON_SPRING,
    DV_SEASON_STATE_ARCHIVED,
    DV_SEASON_STATE_FINISHED,
    DV_SEASON_STATE_OPEN,
    DV_SEASON_STATE_PLANNING,
    DV_SEASON_STATE_RUNNING,
    DV_SEASON_STATES,
    DV_STATE_CHOICES,
    STDGLYPHICON,
)
from apps.common.fields import ChoiceArrayField
from defivelo.templatetags.dv_filters import cantons_abbr


class Season(models.Model):
    created_on = models.DateTimeField(auto_now_add=True)
    year = models.PositiveSmallIntegerField(_("Année"))
    month_start = models.PositiveSmallIntegerField(
        _("Mois de début"), choices=MONTHS.items()
    )
    n_months = models.PositiveSmallIntegerField(
        _("Nombre de mois"),
        default=1,
        validators=[MinValueValidator(1), MaxValueValidator(24)],
    )
    cantons = ChoiceArrayField(
        models.CharField(
            max_length=2,
            choices=sorted(DV_STATE_CHOICES),
        ),
        verbose_name=_("Cantons"),
        default=list,
    )

    leader = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Chargé·e de projet"),
        limit_choices_to={"managedstates__isnull": False},
        on_delete=models.CASCADE,
    )
    state = models.PositiveSmallIntegerField(
        _("État"), choices=DV_SEASON_STATES, default=DV_SEASON_STATE_PLANNING
    )
    availability_reminder_sent_at = models.DateTimeField(
        _("Date d’envoi du rappel de disponibilités"), null=True, blank=True
    )
    history = HistoricalRecords()

    class Meta:
        verbose_name = _p("Singular month", "Mois")
        verbose_name_plural = _p("Plural months", "Mois")
        ordering = [
            "year",
            "month_start",
            "cantons",
        ]

    @cached_property
    def begin(self):
        return date(self.year, self.month_start, 1)

    @cached_property
    def end(self):
        # divmod will put months from 0 to 11, so pad this away
        add_years, final_month_0 = divmod(self.month_start + self.n_months - 1, 12)

        return date(self.year + add_years, final_month_0 + 1, 1) - timedelta(days=1)

    @cached_property
    def state_full(self):
        return dict(DV_SEASON_STATES)[self.state]

    @cached_property
    def state_icon(self):
        icon = ""
        title = self.state_full
        if self.state == DV_SEASON_STATE_PLANNING:
            icon = "calendar"
        elif self.state == DV_SEASON_STATE_OPEN:
            icon = "flash"
        elif self.state == DV_SEASON_STATE_RUNNING:
            icon = "apple"
        elif self.state == DV_SEASON_STATE_FINISHED:
            icon = "floppy-saved"
        elif self.state == DV_SEASON_STATE_ARCHIVED:
            icon = "folder-close"
        if icon:
            return mark_safe(STDGLYPHICON.format(icon=icon, title=title))
        return ""

    @cached_property
    def state_class(self):
        css_class = "default"
        if self.state == DV_SEASON_STATE_PLANNING:
            css_class = "warning"  # Orange
        elif self.state == DV_SEASON_STATE_OPEN:
            css_class = "success"  # Green
        elif self.state == DV_SEASON_STATE_RUNNING:
            css_class = "warning"  # Orange
        elif self.state == DV_SEASON_STATE_FINISHED:
            css_class = "default disabled"  # Black
        elif self.state == DV_SEASON_STATE_ARCHIVED:
            css_class = "default disabled"  # Black
        return css_class

    @cached_property
    def staff_can_update_availability(self):
        return self.state == DV_SEASON_STATE_OPEN

    @cached_property
    def staff_can_see_planning(self):
        return self.state == DV_SEASON_STATE_RUNNING

    @cached_property
    def staff_can_see(self):
        return self.state == DV_SEASON_STATE_RUNNING

    @property
    def can_set_state_open(self):
        return self.state == DV_SEASON_STATE_PLANNING

    @property
    def can_set_state_running(self):
        return self.state in [DV_SEASON_STATE_OPEN]

    @cached_property
    def manager_can_crud(self):
        return self.state != DV_SEASON_STATE_ARCHIVED

    @cached_property
    def coordinator_can_update(self):
        return self.state in [DV_SEASON_STATE_PLANNING, DV_SEASON_STATE_OPEN]

    @cached_property
    def season_full(self):
        # Before November 2020 (DEFIVELO-48), the Season objects were either "Spring" or "Autumn"
        # Check if the current Season=Months matches the legacy objects.
        legacy_season = None
        if self.year < 2021:
            # Historical "Spring" season
            if self.month_start == 1 and self.n_months == DV_SEASON_LAST_SPRING_MONTH:
                legacy_season = DV_SEASON_SPRING
            # Historical "Autumn" season
            if (
                self.month_start == DV_SEASON_LAST_SPRING_MONTH + 1
                and self.n_months == (12 - DV_SEASON_LAST_SPRING_MONTH)
            ):
                legacy_season = DV_SEASON_AUTUMN

        if legacy_season:
            return _("{season} {year}").format(
                season=dict(DV_SEASON_CHOICES)[legacy_season], year=self.year
            )

        if self.begin.year == self.end.year:
            if self.begin.month == self.end.month:
                return _("{month} {year}").format(
                    month=MONTHS[self.begin.month].capitalize(), year=self.year
                )
            return _("{month_begin} - {month_end} {year}").format(
                month_begin=MONTHS[self.begin.month].capitalize(),
                month_end=MONTHS[self.end.month],
                year=self.year,
            )
        return _("{month_begin} {year_begin} - {month_end} {year_end}").format(
            month_begin=MONTHS[self.begin.month].capitalize(),
            month_end=MONTHS[self.end.month],
            year_begin=self.year,
            year_end=self.end.year,
        )

    @property
    def has_availability_incoherences(self):
        for session in self.sessions_with_qualifs:
            if session.has_availability_incoherences:
                return True
        return False

    @property
    def sessions(self):
        from .session import Session

        return Session.objects.filter(
            orga__address_canton__in=self.cantons,
            day__gte=self.begin,
            day__lte=self.end,
        ).prefetch_related(
            "qualifications",
            "qualifications__activity_A",
            "qualifications__activity_B",
            "qualifications__activity_C",
            "qualifications__leader",
            "qualifications__helpers",
            "qualifications__actor",
            "orga",
        )

    @property
    def sessions_by_orga(self):
        return self.sessions.order_by("orga__name", "day", "begin")

    @cached_property
    def sessions_with_qualifs(self):
        if not hasattr(self, "sessions_with_q"):
            self.sessions_with_q = (
                self.sessions.prefetch_related("availability_statuses")
                .annotate(models.Count("qualifications"))
                .filter(qualifications__count__gt=0)
                .order_by("day", "begin", "orga__name")
            )
        return self.sessions_with_q

    @cached_property
    def all_helpers_qs(self):
        User = get_user_model()
        return User.objects.filter(
            Q(qualifs_mon2__session__in=self.sessions_with_qualifs)
            | Q(qualifs_mon1__session__in=self.sessions_with_qualifs)
            | Q(qualifs_actor__session__in=self.sessions_with_qualifs)
        ).distinct()

    @cached_property
    def all_coordinator_qs(self):
        """
        All coordinators with organizations with sessions in this season
        """
        User = get_user_model()
        return User.objects.filter(
            managed_organizations__sessions__in=self.sessions_with_qualifs
        ).distinct()

    def unprivileged_user_can_see(self, user):
        """
        Whether a user can consult this season;
        """
        # All users selected in season can see all of this season's sessions
        # Coordinators can always consult.
        return (
            self.staff_can_see and self.all_helpers_qs.filter(id=user.id).exists()
        ) or self.all_coordinator_qs.filter(id=user.id).exists()

    def get_absolute_url(self):
        return reverse("season-detail", args=[self.pk])

    def save(self, *args, **kwargs):
        """
        Override save to clear some cached properties
        """
        super().save(*args, **kwargs)
        # Clear cached properties
        for cached_prop in [
            "can_set_state_running",
            "staff_can_update_availability",
            "staff_can_see_planning",
            "manager_can_crud",
            "coordinator_can_update",
            "season_full",
        ]:
            try:
                del self.__dict__[cached_prop]
            except KeyError:
                pass

    def desc(self, abbr=False):
        return mark_safe(
            _("{cantons} - {season_full}").format(
                season_full=self.season_full,
                cantons=", ".join(cantons_abbr(self.cantons, abbr)),
            )
        )

    @cached_property
    def desc_abbr(self):
        return self.desc(True)

    def __str__(self):
        return "{desc}{leader}".format(
            desc=self.desc(False),
            leader=" - %s" % self.leader.get_full_name() if self.leader else "",
        )



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.db import models
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from apps.common import STDGLYPHICON
from apps.common.models import Address
from defivelo.templatetags.dv_filters import dv_season_url

User = get_user_model()

ORGASTATUS_UNDEF = 0
ORGASTATUS_ACTIVE = 10
ORGASTATUS_INACTIVE = 30

ORGASTATUS_CHOICES = (
    (ORGASTATUS_UNDEF, "---------"),
    (ORGASTATUS_ACTIVE, _("Actif")),
    (ORGASTATUS_INACTIVE, _("Inactif")),
)


class Organization(Address, models.Model):
    created_on = models.DateTimeField(auto_now_add=True)
    abbr = models.CharField(_("Abréviation"), max_length=16, blank=True)
    name = models.CharField(_("Nom"), max_length=255)
    website = models.URLField(_("Site web"), blank=True)
    status = models.PositiveSmallIntegerField(
        _("Statut"), choices=ORGASTATUS_CHOICES, default=ORGASTATUS_ACTIVE
    )
    comments = models.TextField(_("Remarques"), blank=True)

    coordinator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="managed_organizations",
        verbose_name=_("Coordina·teur·trice"),
        null=True,
        on_delete=models.SET_NULL,
    )

    @cached_property
    def abbr_verb(self):
        return mark_safe(
            '<abbr title="{name}">{abbr}</abbr>'.format(name=self.name, abbr=self.abbr)
        )

    @cached_property
    def abbr_or_name(self) -> str:
        if self.abbr:
            return self.abbr
        return str(self.name).strip()

    @cached_property
    def ifabbr(self):
        return self.abbr_verb if self.abbr else self.name

    @cached_property
    def status_full(self):
        if self.status:
            return dict(ORGASTATUS_CHOICES)[self.status]
        return ""

    def status_icon(self):
        icon = ""
        title = self.status_full
        if self.status == ORGASTATUS_ACTIVE:
            icon = "star"
        elif self.status == ORGASTATUS_INACTIVE:
            icon = "hourglass"
        if icon:
            return mark_safe(STDGLYPHICON.format(icon=icon, title=title))
        return ""

    def status_class(self):
        css_class = "default"
        if self.status == ORGASTATUS_ACTIVE:
            css_class = "success"  # Green
        elif self.status == ORGASTATUS_INACTIVE:
            css_class = "danger"  # Red
        return css_class

    def shortname(self):
        return "{abbr}{city}".format(
            abbr=self.ifabbr,
            city=" (%s)" % self.address_city if self.address_city else "",
        )

    class Meta:
        verbose_name = _("Établissement")
        verbose_name_plural = _("Établissements")
        ordering = ["name"]

    def __str__(self):
        return "{name}{city}".format(
            name=self.name,
            city=" (%s)" % self.address_city if self.address_city else "",
        )

    def get_absolute_url(self):
        return reverse("organization-detail", args=[self.pk])

    def get_state_managers(self):
        return User.objects.filter(managedstates__canton=self.address_canton)

    def notify_new_registrations(self, request):
        for state_manager in self.get_state_managers():
            state_manager.profile.send_mail(
                "%s %s"
                % (
                    settings.EMAIL_SUBJECT_PREFIX,
                    gettext("Nouvelles pré-inscriptions à valider"),
                ),
                render_to_string(
                    "challenge/registration_email_to_state_manager.txt",
                    {
                        "full_name": state_manager.get_full_name(),
                        "url": request.build_absolute_uri(
                            reverse("registration-validate")
                        ),
                        "site_domain": Site.objects.get_current().domain,
                    },
                ),
            )

        messages.success(
            request,
            _(
                "Votre préinscription a été enregistrée et le/la chargé·e de "
                "projet de votre canton a été notifié·e!"
            ),
        )

    def notify_registrations_validated(self, request):
        self.coordinator.profile.send_mail(
            "%s %s"
            % (
                settings.EMAIL_SUBJECT_PREFIX,
                gettext("Vos inscriptions à DEFIVELO"),
            ),
            render_to_string(
                "challenge/registration_email_to_coordinator.txt",
                {
                    "full_name": self.coordinator.get_full_name(),
                    "url": request.build_absolute_uri(dv_season_url()),
                    "site_domain": Site.objects.get_current().domain,
                },
            ),
        )

        messages.success(
            request,
            _(
                "L'inscription pour l'établissement %s est enregistrée, et un "
                "email a été envoyé à la personne coordinatrice." % self
            ),
        )



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015, 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.db.models import Q

from dal_select2.views import Select2QuerySetView
from six import get_unbound_function

from apps.challenge import MAX_MONO1_PER_QUALI
from defivelo.roles import has_permission

from .. import FORMATION_KEYS, FORMATION_M2
from ..models import USERSTATUS_DELETED
from .mixins import ProfileMixin
from .standard import UserProfileFilterSet


class PersonAutocomplete(ProfileMixin, Select2QuerySetView):
    model = get_user_model()
    required_permission = "user_view_list"
    choices = None
    widget_attrs = {
        "data-widget-maximum-values": 1,
    }

    def get_result_label(self, choice):
        return "{name}{ifcanton}".format(
            name=choice.get_full_name(),
            ifcanton=(
                (" (%s)" % choice.profile.affiliation_canton)
                if choice.profile and choice.profile.affiliation_canton
                else ""
            ),
        )

    def get_queryset(self):
        if has_permission(self.request.user, self.required_permission):
            # Remove self.q to bypass Select2QuerySetView' wrong handling
            q = self.q if self.q else None
            del self.q
            qs = super(PersonAutocomplete, self).get_queryset()
            # Only non-deleted
            qs = qs.exclude(profile__status=USERSTATUS_DELETED)
            if q:
                filter_wide = get_unbound_function(UserProfileFilterSet.filter_wide)
                qs = filter_wide(qs, "", q)
            return qs
        else:
            raise PermissionDenied


class AllPersons(PersonAutocomplete):
    pass


class AllCoordinators(PersonAutocomplete):
    def get_queryset(self):
        return super().get_queryset().filter(groups__name="coordinator")


class PersonsRelevantForSessions(PersonAutocomplete):
    def get_queryset(self):
        qs = super(PersonsRelevantForSessions, self).get_queryset()
        qs = qs.filter(
            Q(profile__formation__in=FORMATION_KEYS)
            | Q(profile__actor_for__isnull=False)
        )
        # Filtre par les cantons s'ils existent
        cantons = self.forwarded.get("cantons", [])
        if len(cantons) > 0:
            qs = qs.filter(
                Q(profile__activity_cantons__overlap=cantons)
                | Q(profile__affiliation_canton__in=cantons)
            )
        return qs.distinct()


class Helpers(PersonAutocomplete):
    widget_attrs = {
        "data-widget-maximum-values": MAX_MONO1_PER_QUALI,
    }

    def get_queryset(self):
        qs = super(Helpers, self).get_queryset()
        return qs.filter(profile__formation__in=FORMATION_KEYS)

    def get_result_label(self, choice):
        return "{name} {icon}".format(
            name=choice.get_full_name(), icon=choice.profile.formation_icon()
        )


class Leaders(PersonAutocomplete):
    def get_queryset(self):
        qs = super(Leaders, self).get_queryset()
        return qs.filter(profile__formation=FORMATION_M2)


class Actors(PersonAutocomplete):
    def get_queryset(self):
        qs = super(Actors, self).get_queryset()
        return qs.exclude(profile__actor_for__isnull=True).distinct()



from .timesheets import Timesheet
from .validations import MonthlyCantonalValidation, MonthlyCantonalValidationUrl

__all__ = ["MonthlyCantonalValidation", "MonthlyCantonalValidationUrl", "Timesheet"]



# Generated by Django 1.11.16 on 2018-10-07 14:23
from __future__ import unicode_literals

import datetime

import django.db.models.deletion
import simple_history.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("orga", "0006_auto_20170831_1201"),
        ("common", "0006_unaccented_psql"),
        ("challenge", "0049_helperseasonworkwish"),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalHelperSeasonWorkWish",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("created_on", models.DateTimeField(blank=True, editable=False)),
                (
                    "amount",
                    models.PositiveSmallIntegerField(
                        default=0, verbose_name="Quantit\xe9 d'engagements souhait\xe9s"
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "helper",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "season",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="challenge.Season",
                    ),
                ),
            ],
            options={
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
                "verbose_name": "historical helper season work wish",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalHelperSessionAvailability",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("created_on", models.DateTimeField(blank=True, editable=False)),
                (
                    "availability",
                    models.CharField(
                        choices=[("y", "Oui"), ("i", "Si n\xe9cessaire"), ("n", "Non")],
                        max_length=1,
                        verbose_name="Disponible",
                    ),
                ),
                (
                    "chosen_as",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "Pas choisi"),
                            (1, "Choisi"),
                            (2, "Comme intervenant"),
                            (5, "Moniteur de Secours"),
                            (3, "Moniteur 1"),
                            (4, "Moniteur 2"),
                        ],
                        default=0,
                        verbose_name="S\xe9lectionn\xe9 pour la session comme",
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "helper",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "session",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="challenge.Session",
                    ),
                ),
            ],
            options={
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
                "verbose_name": "historical Disponibilit\xe9 par session",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalQualification",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("created_on", models.DateTimeField(blank=True, editable=False)),
                (
                    "name",
                    models.CharField(max_length=255, verbose_name="Nom de la classe"),
                ),
                (
                    "class_teacher_fullname",
                    models.CharField(
                        blank=True, max_length=512, verbose_name="Enseignant"
                    ),
                ),
                (
                    "class_teacher_natel",
                    models.CharField(
                        blank=True, max_length=13, verbose_name="Natel enseignant"
                    ),
                ),
                (
                    "n_participants",
                    models.PositiveSmallIntegerField(
                        blank=True, null=True, verbose_name="Nombre de participants"
                    ),
                ),
                (
                    "n_bikes",
                    models.PositiveSmallIntegerField(
                        blank=True, null=True, verbose_name="Nombre de v\xe9los"
                    ),
                ),
                (
                    "n_helmets",
                    models.PositiveSmallIntegerField(
                        blank=True, null=True, verbose_name="Nombre de casques"
                    ),
                ),
                ("comments", models.TextField(blank=True, verbose_name="Remarques")),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "activity_A",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="challenge.QualificationActivity",
                    ),
                ),
                (
                    "activity_B",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="challenge.QualificationActivity",
                    ),
                ),
                (
                    "activity_C",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="challenge.QualificationActivity",
                    ),
                ),
                (
                    "actor",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "leader",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "session",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="challenge.Session",
                    ),
                ),
            ],
            options={
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
                "verbose_name": "historical Qualif'",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalQualificationActivity",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("A", "Agilit\xe9"),
                            ("B", "M\xe9canique"),
                            ("C", "Rencontre"),
                        ],
                        max_length=1,
                        verbose_name="Cat\xe9gorie",
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
                "verbose_name": "historical Poste",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalSeason",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("created_on", models.DateTimeField(blank=True, editable=False)),
                ("year", models.PositiveSmallIntegerField(verbose_name="Ann\xe9e")),
                (
                    "season",
                    models.PositiveSmallIntegerField(
                        choices=[(1, "Printemps"), (3, "Automne")],
                        default=1,
                        verbose_name="Saison",
                    ),
                ),
                (
                    "cantons",
                    models.CharField(
                        choices=[
                            (b"AR", "Appenzell Ausserrhoden"),
                            (b"BE", "Berne"),
                            (b"BS", "Basel-Stadt"),
                            (b"FR", "Fribourg"),
                            (b"GE", "Geneva"),
                            (b"JU", "Jura"),
                            (b"LU", "Lucerne"),
                            (b"NE", "Neuchatel"),
                            (b"SG", "St. Gallen"),
                            (b"VD", "Vaud"),
                            (b"VS", "Valais"),
                            ("WS", "Haut-Valais"),
                            (b"ZH", "Zurich"),
                        ],
                        max_length=38,
                        verbose_name="Cantons",
                    ),
                ),
                (
                    "state",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (1, "Planification (invisible)"),
                            (2, "Annonc\xe9e (rentr\xe9e des disponibilit\xe9s)"),
                            (
                                3,
                                "En cours (corrections que par charg\xe9\xb7e\xb7s de projet)",
                            ),
                            (4, "Termin\xe9e"),
                            (5, "Archiv\xe9e"),
                        ],
                        default=1,
                        verbose_name="\xc9tat",
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "leader",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
                "verbose_name": "historical Saison",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalSession",
            fields=[
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "address_street",
                    models.CharField(blank=True, max_length=255, verbose_name="Rue"),
                ),
                (
                    "address_no",
                    models.CharField(blank=True, max_length=8, verbose_name="N\xb0"),
                ),
                (
                    "address_additional",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        verbose_name="Compl\xe9ment d'adresse",
                    ),
                ),
                (
                    "address_zip",
                    models.CharField(blank=True, max_length=4, verbose_name="NPA"),
                ),
                (
                    "address_city",
                    models.CharField(blank=True, max_length=64, verbose_name="Ville"),
                ),
                (
                    "address_canton",
                    models.CharField(blank=True, max_length=2, verbose_name="Canton"),
                ),
                ("created_on", models.DateTimeField(blank=True, editable=False)),
                ("day", models.DateField(blank=True, verbose_name="Date")),
                (
                    "begin",
                    models.TimeField(blank=True, null=True, verbose_name="D\xe9but"),
                ),
                (
                    "duration",
                    models.DurationField(
                        default=datetime.timedelta(0, 10800), verbose_name="Dur\xe9e"
                    ),
                ),
                (
                    "place",
                    models.CharField(
                        blank=True, max_length=512, verbose_name="Lieu de la Qualif'"
                    ),
                ),
                (
                    "fallback_plan",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("A", "Programme d\xe9luge"),
                            ("B", "Annulation"),
                            ("C", "Report \u2026"),
                            ("D", "Autre \u2026"),
                        ],
                        max_length=1,
                        verbose_name="Mauvais temps",
                    ),
                ),
                (
                    "helpers_time",
                    models.TimeField(
                        blank=True,
                        null=True,
                        verbose_name="Heure rendez-vous moniteurs",
                    ),
                ),
                (
                    "helpers_place",
                    models.CharField(
                        blank=True,
                        max_length=512,
                        verbose_name="Lieu rendez-vous moniteurs",
                    ),
                ),
                (
                    "apples",
                    models.CharField(blank=True, max_length=512, verbose_name="Pommes"),
                ),
                (
                    "bikes_concept",
                    models.CharField(
                        blank=True, max_length=512, verbose_name="Logistique v\xe9los"
                    ),
                ),
                (
                    "bikes_phone",
                    models.CharField(
                        blank=True,
                        max_length=13,
                        verbose_name="N\xb0 de contact v\xe9los",
                    ),
                ),
                ("comments", models.TextField(blank=True, verbose_name="Remarques")),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "address_ptr",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="common.Address",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "orga",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="orga.Organization",
                    ),
                ),
                (
                    "superleader",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
                "verbose_name": "historical Session",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]
